#!/usr/bin/python

# Copyright (c) James Laska (jlaska@redhat.com)
#
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: redhat_subscription
short_description: Manage registration and subscriptions to RHSM using C(subscription-manager)
description:
  - Manage registration and subscription to the Red Hat Subscription Management entitlement platform using the C(subscription-manager)
    command, registering using D-Bus if possible.
author: "Barnaby Court (@barnabycourt)"
notes:
  - 'The module tries to use the D-Bus C(rhsm) service (part of C(subscription-manager)) to register, starting from community.general
    6.5.0: this is done so credentials (username, password, activation keys) can be passed to C(rhsm) in a secure way. C(subscription-manager)
    itself gets credentials only as arguments of command line parameters, which is I(not) secure, as they can be easily stolen
    by checking the process listing on the system. Due to limitations of the D-Bus interface of C(rhsm), the module does I(not)
    use D-Bus for registration when trying either to register using O(token), or when specifying O(environment), or when the
    system is old (typically RHEL 7 older than 7.4, RHEL 6, and older).'
  - In order to register a system, subscription-manager requires either a username and password, or an activationkey and an
    Organization ID.
  - Since 2.5 values for O(server_hostname), O(server_insecure), O(rhsm_baseurl), O(server_proxy_hostname), O(server_proxy_port),
    O(server_proxy_user) and O(server_proxy_password) are no longer taken from the C(/etc/rhsm/rhsm.conf) config file and
    default to V(null).
  - It is possible to interact with C(subscription-manager) only as root, so root permissions are required to successfully
    run this module.
  - Since community.general 6.5.0, credentials (that is, O(username) and O(password), O(activationkey), or O(token)) are needed
    only in case the system is not registered, or O(force_register) is specified; this makes it possible to use the module
    to tweak an already registered system, for example attaching pools to it (using O(pool_ids)), and modifying the C(syspurpose)
    attributes (using O(syspurpose)).
requirements:
  - subscription-manager
  - Optionally the C(dbus) Python library; this is usually included in the OS as it is used by C(subscription-manager).
extends_documentation_fragment:
  - community.general._attributes
attributes:
  check_mode:
    support: none
  diff_mode:
    support: none
options:
  state:
    description:
      - Whether to register and subscribe (V(present)), or unregister (V(absent)) a system.
    choices: ["present", "absent"]
    default: "present"
    type: str
  username:
    description:
      - Access.redhat.com or Red Hat Satellite or Katello username.
    type: str
  password:
    description:
      - Access.redhat.com or Red Hat Satellite or Katello password.
    type: str
  token:
    description:
      - Sso.redhat.com API access token.
    type: str
    version_added: 6.3.0
  server_hostname:
    description:
      - Specify an alternative Red Hat Subscription Management or Red Hat Satellite or Katello server.
    type: str
  server_insecure:
    description:
      - Enable or disable https server certificate verification when connecting to O(server_hostname).
    type: str
  server_prefix:
    description:
      - Specify the prefix when registering to the Red Hat Subscription Management or Red Hat Satellite or Katello server.
    type: str
    version_added: 3.3.0
  server_port:
    description:
      - Specify the port when registering to the Red Hat Subscription Management or Red Hat Satellite or Katello server.
    type: str
    version_added: 3.3.0
  rhsm_baseurl:
    description:
      - Specify CDN baseurl.
    type: str
  rhsm_repo_ca_cert:
    description:
      - Specify an alternative location for a CA certificate for CDN.
    type: str
  server_proxy_hostname:
    description:
      - Specify an HTTP proxy hostname.
    type: str
  server_proxy_scheme:
    description:
      - Specify an HTTP proxy scheme, for example V(http) or V(https).
    type: str
    version_added: 6.2.0
  server_proxy_port:
    description:
      - Specify an HTTP proxy port.
    type: str
  server_proxy_user:
    description:
      - Specify a user for HTTP proxy with basic authentication.
    type: str
  server_proxy_password:
    description:
      - Specify a password for HTTP proxy with basic authentication.
    type: str
  auto_attach:
    description:
      - Upon successful registration, auto-consume available subscriptions.
      - Please note that the alias O(ignore:autosubscribe) was removed in community.general 9.0.0.
    type: bool
  activationkey:
    description:
      - Supply an activation key for use with registration.
    type: str
  org_id:
    description:
      - Organization ID to use in conjunction with activationkey.
    type: str
  environment:
    description:
      - Register with a specific environment in the destination org. Used with Red Hat Satellite or Katello.
    type: str
  pool_ids:
    description:
      - Specify subscription pool IDs to consume.
      - 'A pool ID may be specified as a C(string) - just the pool ID (for example V(0123456789abcdef0123456789abcdef)), or
        as a C(dict) with the pool ID as the key, and a quantity as the value (for example V(0123456789abcdef0123456789abcdef:
        2). If the quantity is provided, it is used to consume multiple entitlements from a pool (the pool must support this).'
    default: []
    type: list
    elements: raw
  consumer_type:
    description:
      - The type of unit to register, defaults to system.
    type: str
  consumer_name:
    description:
      - Name of the system to register, defaults to the hostname.
    type: str
  consumer_id:
    description:
      - References an existing consumer ID to resume using a previous registration for this system. If the system's identity
        certificate is lost or corrupted, this option allows it to resume using its previous identity and subscriptions. The
        default is to not specify a consumer ID so a new ID is created.
    type: str
  force_register:
    description:
      - Register the system even if it is already registered.
    type: bool
    default: false
  release:
    description:
      - Set a release version.
    type: str
  syspurpose:
    description:
      - Set syspurpose attributes in file C(/etc/rhsm/syspurpose/syspurpose.json) and synchronize these attributes with RHSM
        server. Syspurpose attributes help attach the most appropriate subscriptions to the system automatically. When C(syspurpose.json)
        file already contains some attributes, then new attributes overwrite existing attributes. When some attribute is not
        listed in the new list of attributes, the existing attribute is removed from C(syspurpose.json) file. Unknown attributes
        are ignored.
    type: dict
    suboptions:
      usage:
        description: Syspurpose attribute usage.
        type: str
      role:
        description: Syspurpose attribute role.
        type: str
      service_level_agreement:
        description: Syspurpose attribute service_level_agreement.
        type: str
      addons:
        description: Syspurpose attribute addons.
        type: list
        elements: str
      sync:
        description:
          - When this option is V(true), then syspurpose attributes are synchronized with RHSM server immediately. When this
            option is V(false), then syspurpose attributes are synchronized with RHSM server by rhsmcertd daemon.
        type: bool
        default: false
"""

EXAMPLES = r"""
- name: Register as user (joe_user) with password (somepass) and auto-subscribe to available content.
  community.general.redhat_subscription:
    state: present
    username: joe_user
    password: somepass
    auto_attach: true

- name: Same as above but subscribe to a specific pool by ID.
  community.general.redhat_subscription:
    state: present
    username: joe_user
    password: somepass
    pool_ids: 0123456789abcdef0123456789abcdef

- name: Register and subscribe to multiple pools.
  community.general.redhat_subscription:
    state: present
    username: joe_user
    password: somepass
    pool_ids:
      - 0123456789abcdef0123456789abcdef
      - 1123456789abcdef0123456789abcdef

- name: Same as above but consume multiple entitlements.
  community.general.redhat_subscription:
    state: present
    username: joe_user
    password: somepass
    pool_ids:
      - 0123456789abcdef0123456789abcdef: 2
      - 1123456789abcdef0123456789abcdef: 4

- name: Register and pull existing system data.
  community.general.redhat_subscription:
    state: present
    username: joe_user
    password: somepass
    consumer_id: xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx

- name: Register as user credentials into given environment (against Red Hat Satellite or Katello), and auto-subscribe.
  community.general.redhat_subscription:
    state: present
    username: joe_user
    password: somepass
    environment: Library
    auto_attach: true

- name: Register as user (joe_user) with password (somepass) and a specific release
  community.general.redhat_subscription:
    state: present
    username: joe_user
    password: somepass
    release: 7.4

- name: Register as user (joe_user) with password (somepass), set syspurpose attributes and synchronize them with server
  community.general.redhat_subscription:
    state: present
    username: joe_user
    password: somepass
    auto_attach: true
    syspurpose:
      usage: "Production"
      role: "Red Hat Enterprise Server"
      service_level_agreement: "Premium"
      addons:
        - addon1
        - addon2
      sync: true
"""

RETURN = r"""
subscribed_pool_ids:
  description: List of pool IDs to which system is now subscribed.
  returned: success
  type: dict
  sample: {"8a85f9815ab905d3015ab928c7005de4": "1"}
"""

import configparser
import json
import re
import shutil
import tempfile
from os import getuid, unlink
from os.path import isfile

from ansible.module_utils import distro
from ansible.module_utils.basic import AnsibleModule
from ansible.module_utils.common.text.converters import to_native

SUBMAN_CMD = None


class Rhsm:
    REDHAT_REPO = "/etc/yum.repos.d/redhat.repo"

    def __init__(self, module):
        self.module = module

    def update_plugin_conf(self, plugin, enabled=True):
        plugin_conf = f"/etc/yum/pluginconf.d/{plugin}.conf"

        if isfile(plugin_conf):
            tmpfd, tmpfile = tempfile.mkstemp()
            shutil.copy2(plugin_conf, tmpfile)
            cfg = configparser.ConfigParser()
            cfg.read([tmpfile])

            if enabled:
                cfg.set("main", "enabled", "1")
            else:
                cfg.set("main", "enabled", "0")

            with open(tmpfile, "w+") as fd:
                cfg.write(fd)
            self.module.atomic_move(tmpfile, plugin_conf)

    def enable(self):
        """
        Enable the system to receive updates from subscription-manager.
        This involves updating affected yum plugins and removing any
        conflicting yum repositories.
        """
        # Remove any existing redhat.repo
        if isfile(self.REDHAT_REPO):
            unlink(self.REDHAT_REPO)
        self.update_plugin_conf("rhnplugin", False)
        self.update_plugin_conf("subscription-manager", True)

    def configure(self, **kwargs):
        """
        Configure the system as directed for registration with RHSM
        Raises:
          * Exception - if error occurs while running command
        """

        args = [SUBMAN_CMD, "config"]

        # Pass supplied **kwargs as parameters to subscription-manager.  Ignore
        # non-configuration parameters and replace '_' with '.'.  For example,
        # 'server_hostname' becomes '--server.hostname'.
        options = []
        for k, v in sorted(kwargs.items()):
            if re.search(r"^(server|rhsm)_", k) and v is not None:
                options.append(f"--{k.replace('_', '.', 1)}={v}")

        # When there is nothing to configure, then it is not necessary
        # to run config command, because it only returns current
        # content of current configuration file
        if len(options) == 0:
            return

        args.extend(options)

        self.module.run_command(args, check_rc=True)

    @property
    def is_registered(self):
        """
        Determine whether the current system
        Returns:
          * Boolean - whether the current system is currently registered to
                      RHSM.
        """

        args = [SUBMAN_CMD, "identity"]
        rc, stdout, stderr = self.module.run_command(args, check_rc=False)
        return rc == 0

    def _has_dbus_interface(self):
        """
        Checks whether subscription-manager has a D-Bus interface.

        :returns: bool -- whether subscription-manager has a D-Bus interface.
        """

        def str2int(s, default=0):
            try:
                return int(s)
            except ValueError:
                return default

        distro_id = distro.id()
        distro_version = tuple(str2int(p) for p in distro.version_parts())

        # subscription-manager in any supported Fedora version has the interface.
        if distro_id == "fedora":
            return True
        # Any other distro: assume it is EL;
        # the D-Bus interface was added to subscription-manager in RHEL 7.4.
        return (distro_version[0] == 7 and distro_version[1] >= 4) or distro_version[0] >= 8

    def _can_connect_to_dbus(self):
        """
        Checks whether it is possible to connect to the system D-Bus bus.

        :returns: bool -- whether it is possible to connect to the system D-Bus bus.
        """

        try:
            # Technically speaking, subscription-manager uses dbus-python
            # as D-Bus library, so this ought to work; better be safe than
            # sorry, I guess...
            import dbus
        except ImportError:
            self.module.debug("dbus Python module not available, will use CLI")
            return False

        try:
            bus = dbus.SystemBus()
            msg = dbus.lowlevel.SignalMessage("/", "com.example", "test")
            bus.send_message(msg)
            bus.flush()

        except dbus.exceptions.DBusException as e:
            self.module.debug(f"Failed to connect to system D-Bus bus, will use CLI: {e}")
            return False

        self.module.debug("Verified system D-Bus bus as usable")
        return True

    def register(
        self,
        was_registered,
        username,
        password,
        token,
        auto_attach,
        activationkey,
        org_id,
        consumer_type,
        consumer_name,
        consumer_id,
        force_register,
        environment,
        release,
    ):
        """
        Register the current system to the provided RHSM or Red Hat Satellite
        or Katello server

        Raises:
          * Exception - if any error occurs during the registration
        """
        # There is no support for token-based registration in the D-Bus API
        # of rhsm, so always use the CLI in that case;
        # also, since the specified environments are names, and the D-Bus APIs
        # require IDs for the environments, use the CLI also in that case
        if not token and not environment and self._has_dbus_interface() and self._can_connect_to_dbus():
            self._register_using_dbus(
                was_registered,
                username,
                password,
                auto_attach,
                activationkey,
                org_id,
                consumer_type,
                consumer_name,
                consumer_id,
                force_register,
                environment,
                release,
            )
            return
        self._register_using_cli(
            username,
            password,
            token,
            auto_attach,
            activationkey,
            org_id,
            consumer_type,
            consumer_name,
            consumer_id,
            force_register,
            environment,
            release,
        )

    def _register_using_cli(
        self,
        username,
        password,
        token,
        auto_attach,
        activationkey,
        org_id,
        consumer_type,
        consumer_name,
        consumer_id,
        force_register,
        environment,
        release,
    ):
        """
        Register using the 'subscription-manager' command

        Raises:
          * Exception - if error occurs while running command
        """
        args = [SUBMAN_CMD, "register"]

        # Generate command arguments
        if force_register:
            args.extend(["--force"])

        if org_id:
            args.extend(["--org", org_id])

        if auto_attach:
            args.append("--auto-attach")

        if consumer_type:
            args.extend(["--type", consumer_type])

        if consumer_name:
            args.extend(["--name", consumer_name])

        if consumer_id:
            args.extend(["--consumerid", consumer_id])

        if environment:
            args.extend(["--environment", environment])

        if activationkey:
            args.extend(["--activationkey", activationkey])
        elif token:
            args.extend(["--token", token])
        else:
            if username:
                args.extend(["--username", username])
            if password:
                args.extend(["--password", password])

        if release:
            args.extend(["--release", release])

        rc, stderr, stdout = self.module.run_command(args, check_rc=True, expand_user_and_vars=False)

    def _register_using_dbus(
        self,
        was_registered,
        username,
        password,
        auto_attach,
        activationkey,
        org_id,
        consumer_type,
        consumer_name,
        consumer_id,
        force_register,
        environment,
        release,
    ):
        """
        Register using D-Bus (connecting to the rhsm service)

        Raises:
          * Exception - if error occurs during the D-Bus communication
        """
        import dbus

        SUBSCRIPTION_MANAGER_LOCALE = "C"
        # Seconds to wait for Registration to complete over DBus;
        # 10 minutes should be a pretty generous timeout.
        REGISTRATION_TIMEOUT = 600

        def str2int(s, default=0):
            try:
                return int(s)
            except ValueError:
                return default

        distro_id = distro.id()
        distro_version_parts = distro.version_parts()
        distro_version = tuple(str2int(p) for p in distro_version_parts)

        # Stop the rhsm service when using systemd (which means Fedora or
        # RHEL 7+): this is because the service may not use new configuration bits
        # - with subscription-manager < 1.26.5-1 (in RHEL < 8.2);
        #   fixed later by https://github.com/candlepin/subscription-manager/pull/2175
        # - sporadically: https://bugzilla.redhat.com/show_bug.cgi?id=2049296
        if distro_id == "fedora" or distro_version[0] >= 7:
            cmd = ["systemctl", "stop", "rhsm"]
            self.module.run_command(cmd, check_rc=True, expand_user_and_vars=False)

        # While there is a 'force' options for the registration, it is actually
        # not implemented (and thus it does not work)
        # - in RHEL 7 and earlier
        # - in RHEL 8 before 8.8: https://bugzilla.redhat.com/show_bug.cgi?id=2118486
        # - in RHEL 9 before 9.2: https://bugzilla.redhat.com/show_bug.cgi?id=2121350
        # Hence, use it only when implemented, manually unregistering otherwise.
        # Match it on RHEL, since we know about it; other distributions
        # will need their own logic.
        dbus_force_option_works = False
        if distro_id == "rhel" and (
            (distro_version[0] == 8 and distro_version[1] >= 8)
            or (distro_version[0] == 9 and distro_version[1] >= 2)
            or distro_version[0] > 9
        ):
            dbus_force_option_works = True
        # We need to use the 'enable_content' D-Bus option to ensure that
        # content is enabled; sadly the option is available depending on the
        # version of the distro, and also depending on which API/method is used
        # for registration.
        dbus_has_enable_content_option = False
        if activationkey:

            def supports_enable_content_for_activation_keys():
                # subscription-manager in Fedora >= 41 has the new option.
                if distro_id == "fedora" and distro_version[0] >= 41:
                    return True
                # Assume EL distros here.
                if distro_version[0] >= 10:  # noqa: SIM103
                    return True
                return False

            dbus_has_enable_content_option = supports_enable_content_for_activation_keys()
        else:

            def supports_enable_content_for_credentials():
                # subscription-manager in any supported Fedora version
                # has the new option.
                if distro_id == "fedora":
                    return True
                # Check for RHEL 8 >= 8.6, or RHEL >= 9.
                if distro_id == "rhel" and (
                    (distro_version[0] == 8 and distro_version[1] >= 6) or distro_version[0] >= 9
                ):
                    return True
                # CentOS: similar checks as for RHEL, with one extra bit:
                # if the 2nd part of the version is empty, it means it is
                # CentOS Stream, and thus we can assume it has the latest
                # version of subscription-manager.
                if distro_id == "centos" and (  # noqa: SIM103
                    (distro_version[0] == 8 and (distro_version[1] >= 6 or distro_version_parts[1] == ""))
                    or distro_version[0] >= 9
                ):
                    return True
                # Unknown or old distro: assume it does not support
                # the new option.
                return False

            dbus_has_enable_content_option = supports_enable_content_for_credentials()

        if force_register and not dbus_force_option_works and was_registered:
            self.unregister()

        register_opts = {}
        if consumer_type:
            # The option for the consumer type used to be 'type' in versions
            # of RHEL before 9 & in RHEL 9 before 9.2, and then it changed to
            # 'consumer_type'; since the Register*() D-Bus functions reject
            # unknown options, we have to pass the right option depending on
            # the version -- funky.
            def supports_option_consumer_type():
                # subscription-manager in any supported Fedora version
                # has the new option.
                if distro_id == "fedora":
                    return True
                # Check for RHEL 9 >= 9.2, or RHEL >= 10.
                if distro_id == "rhel" and (
                    (distro_version[0] == 9 and distro_version[1] >= 2) or distro_version[0] >= 10
                ):
                    return True
                # CentOS: since the change was only done in EL 9, then there is
                # only CentOS Stream for 9, and thus we can assume it has the
                # latest version of subscription-manager.
                if distro_id == "centos" and distro_version[0] >= 9:  # noqa: SIM103
                    return True
                # Unknown or old distro: assume it does not support
                # the new option.
                return False

            consumer_type_key = "type"
            if supports_option_consumer_type():
                consumer_type_key = "consumer_type"
            register_opts[consumer_type_key] = consumer_type
        if consumer_name:
            register_opts["name"] = consumer_name
        if consumer_id:
            register_opts["consumerid"] = consumer_id
        if environment:
            # The option for environments used to be 'environment' in versions
            # of RHEL before 8.6, and then it changed to 'environments'; since
            # the Register*() D-Bus functions reject unknown options, we have
            # to pass the right option depending on the version -- funky.
            def supports_option_environments():
                # subscription-manager in any supported Fedora version
                # has the new option.
                if distro_id == "fedora":
                    return True
                # Check for RHEL 8 >= 8.6, or RHEL >= 9.
                if distro_id == "rhel" and (
                    (distro_version[0] == 8 and distro_version[1] >= 6) or distro_version[0] >= 9
                ):
                    return True
                # CentOS: similar checks as for RHEL, with one extra bit:
                # if the 2nd part of the version is empty, it means it is
                # CentOS Stream, and thus we can assume it has the latest
                # version of subscription-manager.
                if distro_id == "centos" and (  # noqa: SIM103
                    (distro_version[0] == 8 and (distro_version[1] >= 6 or distro_version_parts[1] == ""))
                    or distro_version[0] >= 9
                ):
                    return True
                # Unknown or old distro: assume it does not support
                # the new option.
                return False

            environment_key = "environment"
            if supports_option_environments():
                environment_key = "environments"
            register_opts[environment_key] = environment
        if force_register and dbus_force_option_works and was_registered:
            register_opts["force"] = True
        if dbus_has_enable_content_option:
            register_opts["enable_content"] = "1"
        # Wrap it as proper D-Bus dict
        register_opts = dbus.Dictionary(register_opts, signature="sv", variant_level=1)

        connection_opts = {}
        # Wrap it as proper D-Bus dict
        connection_opts = dbus.Dictionary(connection_opts, signature="sv", variant_level=1)

        bus = dbus.SystemBus()
        register_server = bus.get_object("com.redhat.RHSM1", "/com/redhat/RHSM1/RegisterServer")
        address = register_server.Start(
            SUBSCRIPTION_MANAGER_LOCALE,
            dbus_interface="com.redhat.RHSM1.RegisterServer",
        )

        try:
            # Use the private bus to register the system
            self.module.debug("Connecting to the private DBus")
            private_bus = dbus.connection.Connection(address)

            try:
                if activationkey:
                    args = (
                        org_id,
                        [activationkey],
                        register_opts,
                        connection_opts,
                        SUBSCRIPTION_MANAGER_LOCALE,
                    )
                    private_bus.call_blocking(
                        "com.redhat.RHSM1",
                        "/com/redhat/RHSM1/Register",
                        "com.redhat.RHSM1.Register",
                        "RegisterWithActivationKeys",
                        "sasa{sv}a{sv}s",
                        args,
                        timeout=REGISTRATION_TIMEOUT,
                    )
                else:
                    args = (
                        org_id or "",
                        username,
                        password,
                        register_opts,
                        connection_opts,
                        SUBSCRIPTION_MANAGER_LOCALE,
                    )
                    private_bus.call_blocking(
                        "com.redhat.RHSM1",
                        "/com/redhat/RHSM1/Register",
                        "com.redhat.RHSM1.Register",
                        "Register",
                        "sssa{sv}a{sv}s",
                        args,
                        timeout=REGISTRATION_TIMEOUT,
                    )

            except dbus.exceptions.DBusException as e:
                # Sometimes we get NoReply but the registration has succeeded.
                # Check the registration status before deciding if this is an error.
                if e.get_dbus_name() == "org.freedesktop.DBus.Error.NoReply":
                    if not self.is_registered():
                        # Host is not registered so re-raise the error
                        raise
                else:
                    raise
                # Host was registered so continue
        finally:
            # Always shut down the private bus
            self.module.debug("Shutting down private DBus instance")
            register_server.Stop(
                SUBSCRIPTION_MANAGER_LOCALE,
                dbus_interface="com.redhat.RHSM1.RegisterServer",
            )

        # Make sure to refresh all the local data: this will fetch all the
        # certificates, update redhat.repo, etc.
        self.module.run_command([SUBMAN_CMD, "refresh"], check_rc=True, expand_user_and_vars=False)

        if auto_attach:
            args = [SUBMAN_CMD, "attach", "--auto"]
            self.module.run_command(args, check_rc=True, expand_user_and_vars=False)

        # There is no support for setting the release via D-Bus, so invoke
        # the CLI for this.
        if release:
            args = [SUBMAN_CMD, "release", "--set", release]
            self.module.run_command(args, check_rc=True, expand_user_and_vars=False)

    def unsubscribe(self, serials=None):
        """
        Unsubscribe a system from subscribed channels
        Args:
          serials(list or None): list of serials to unsubscribe. If
                                 serials is none or an empty list, then
                                 all subscribed channels will be removed.
        Raises:
          * Exception - if error occurs while running command
        """
        items = []
        if serials is not None and serials:
            items = [f"--serial={s}" for s in serials]
        if serials is None:
            items = ["--all"]

        if items:
            args = [SUBMAN_CMD, "remove"] + items
            rc, stderr, stdout = self.module.run_command(args, check_rc=True)
        return serials

    def unregister(self):
        """
        Unregister a currently registered system
        Raises:
          * Exception - if error occurs while running command
        """
        args = [SUBMAN_CMD, "unregister"]
        rc, stderr, stdout = self.module.run_command(args, check_rc=True)
        self.update_plugin_conf("rhnplugin", False)
        self.update_plugin_conf("subscription-manager", False)

    def subscribe_by_pool_ids(self, pool_ids):
        """
        Try to subscribe to the list of pool IDs
        """
        available_pools = RhsmPools(self.module)

        available_pool_ids = [p.get_pool_id() for p in available_pools]

        for pool_id, quantity in sorted(pool_ids.items()):
            if pool_id in available_pool_ids:
                args = [SUBMAN_CMD, "attach", "--pool", pool_id]
                if quantity is not None:
                    args.extend(["--quantity", to_native(quantity)])
                rc, stderr, stdout = self.module.run_command(args, check_rc=True)
            else:
                self.module.fail_json(msg=f"Pool ID: {pool_id} not in list of available pools")
        return pool_ids

    def update_subscriptions_by_pool_ids(self, pool_ids):
        changed = False
        consumed_pools = RhsmPools(self.module, consumed=True)

        existing_pools = {}
        serials_to_remove = []
        for p in consumed_pools:
            pool_id = p.get_pool_id()
            quantity_used = p.get_quantity_used()
            existing_pools[pool_id] = quantity_used

            quantity = pool_ids.get(pool_id, 0)
            if quantity is not None and quantity != quantity_used:
                serials_to_remove.append(p.Serial)

        serials = self.unsubscribe(serials=serials_to_remove)

        missing_pools = {}
        for pool_id, quantity in sorted(pool_ids.items()):
            quantity_used = existing_pools.get(pool_id, 0)
            if quantity is None and quantity_used == 0 or quantity not in (None, 0, quantity_used):
                missing_pools[pool_id] = quantity

        self.subscribe_by_pool_ids(missing_pools)

        if missing_pools or serials:
            changed = True
        return {"changed": changed, "subscribed_pool_ids": list(missing_pools.keys()), "unsubscribed_serials": serials}

    def sync_syspurpose(self):
        """
        Try to synchronize syspurpose attributes with server
        """
        args = [SUBMAN_CMD, "status"]
        rc, stdout, stderr = self.module.run_command(args, check_rc=False)


class RhsmPool:
    """
    Convenience class for housing subscription information
    """

    def __init__(self, module, **kwargs):
        self.module = module
        for k, v in kwargs.items():
            setattr(self, k, v)

    def __str__(self):
        return str(self.__getattribute__("_name"))

    def get_pool_id(self):
        return getattr(self, "PoolId", self.PoolID)

    def get_quantity_used(self):
        return int(self.QuantityUsed)

    def subscribe(self):
        args = f"subscription-manager attach --pool {self.get_pool_id()}"
        rc, stdout, stderr = self.module.run_command(args, check_rc=True)
        return rc == 0


class RhsmPools:
    """
    This class is used for manipulating pools subscriptions with RHSM
    """

    def __init__(self, module, consumed=False):
        self.module = module
        self.products = self._load_product_list(consumed)

    def __iter__(self):
        return self.products.__iter__()

    def _load_product_list(self, consumed=False):
        """
        Loads list of all available or consumed pools for system in data structure

        Args:
            consumed(bool): if True list consumed  pools, else list available pools (default False)
        """
        args = "subscription-manager list"
        if consumed:
            args += " --consumed"
        else:
            args += " --available"
        lang_env = dict(LANGUAGE="C", LC_ALL="C")
        rc, stdout, stderr = self.module.run_command(args, check_rc=True, environ_update=lang_env)

        products = []
        for line in stdout.split("\n"):
            # Remove leading+trailing whitespace
            line = line.strip()
            # An empty line implies the end of a output group
            if len(line) == 0:
                continue
            # If a colon ':' is found, parse
            elif ":" in line:
                (key, value) = line.split(":", 1)
                key = key.strip().replace(" ", "")  # To unify
                value = value.strip()
                if key in ["ProductName", "SubscriptionName"]:
                    # Remember the name for later processing
                    products.append(RhsmPool(self.module, _name=value, key=value))
                elif products:
                    # Associate value with most recently recorded product
                    products[-1].__setattr__(key, value)
                # FIXME - log some warning?
                # else:
                # warnings.warn("Unhandled subscription key/value: %s/%s" % (key,value))
        return products

    def filter_pools(self, regexp="^$"):
        """
        Return a list of RhsmPools whose pool id matches the provided regular expression
        """
        r = re.compile(regexp)
        for product in self.products:
            if r.search(product.get_pool_id()):
                yield product

    def filter_products(self, regexp="^$"):
        """
        Return a list of RhsmPools whose product name matches the provided regular expression
        """
        r = re.compile(regexp)
        for product in self.products:
            if r.search(product._name):
                yield product


class SysPurpose:
    """
    This class is used for reading and writing to syspurpose.json file
    """

    SYSPURPOSE_FILE_PATH = "/etc/rhsm/syspurpose/syspurpose.json"

    ALLOWED_ATTRIBUTES = ["role", "usage", "service_level_agreement", "addons"]

    def __init__(self, path=None):
        """
        Initialize class used for reading syspurpose json file
        """
        self.path = path or self.SYSPURPOSE_FILE_PATH

    def update_syspurpose(self, new_syspurpose):
        """
        Try to update current syspurpose with new attributes from new_syspurpose
        """
        syspurpose = {}
        syspurpose_changed = False
        for key, value in new_syspurpose.items():
            if key in self.ALLOWED_ATTRIBUTES:
                if value is not None:
                    syspurpose[key] = value
            elif key == "sync":
                pass
            else:
                raise KeyError(f"Attribute: {key} not in list of allowed attributes: {self.ALLOWED_ATTRIBUTES}")
        current_syspurpose = self._read_syspurpose()
        if current_syspurpose != syspurpose:
            syspurpose_changed = True
        # Update current syspurpose with new values
        current_syspurpose.update(syspurpose)
        # When some key is not listed in new syspurpose, then delete it from current syspurpose
        # and ignore custom attributes created by user (e.g. "foo": "bar")
        for key in list(current_syspurpose):
            if key in self.ALLOWED_ATTRIBUTES and key not in syspurpose:
                del current_syspurpose[key]
        self._write_syspurpose(current_syspurpose)
        return syspurpose_changed

    def _write_syspurpose(self, new_syspurpose):
        """
        This function tries to update current new_syspurpose attributes to
        json file.
        """
        with open(self.path, "w") as fp:
            fp.write(json.dumps(new_syspurpose, indent=2, ensure_ascii=False, sort_keys=True))

    def _read_syspurpose(self):
        """
        Read current syspurpuse from json file.
        """
        current_syspurpose = {}
        try:
            with open(self.path) as fp:
                content = fp.read()
        except OSError:
            pass
        else:
            current_syspurpose = json.loads(content)
        return current_syspurpose


def main():
    # Note: the default values for parameters are:
    # 'type': 'str', 'default': None, 'required': False
    # So there is no need to repeat these values for each parameter.
    module = AnsibleModule(
        argument_spec={
            "state": {"default": "present", "choices": ["present", "absent"]},
            "username": {},
            "password": {"no_log": True},
            "token": {"no_log": True},
            "server_hostname": {},
            "server_insecure": {},
            "server_prefix": {},
            "server_port": {},
            "rhsm_baseurl": {},
            "rhsm_repo_ca_cert": {},
            "auto_attach": {"type": "bool"},
            "activationkey": {"no_log": True},
            "org_id": {},
            "environment": {},
            "pool_ids": {"default": [], "type": "list", "elements": "raw"},
            "consumer_type": {},
            "consumer_name": {},
            "consumer_id": {},
            "force_register": {"default": False, "type": "bool"},
            "server_proxy_hostname": {},
            "server_proxy_scheme": {},
            "server_proxy_port": {},
            "server_proxy_user": {},
            "server_proxy_password": {"no_log": True},
            "release": {},
            "syspurpose": {
                "type": "dict",
                "options": {
                    "role": {},
                    "usage": {},
                    "service_level_agreement": {},
                    "addons": {"type": "list", "elements": "str"},
                    "sync": {"type": "bool", "default": False},
                },
            },
        },
        required_together=[
            ["username", "password"],
            ["server_proxy_hostname", "server_proxy_port"],
            ["server_proxy_user", "server_proxy_password"],
        ],
        mutually_exclusive=[
            ["activationkey", "username"],
            ["activationkey", "token"],
            ["token", "username"],
            ["activationkey", "consumer_id"],
            ["activationkey", "environment"],
            ["activationkey", "auto_attach"],
        ],
        required_if=[["force_register", True, ["username", "activationkey", "token"], True]],
    )

    if getuid() != 0:
        module.fail_json(msg="Interacting with subscription-manager requires root permissions ('become: true')")

    # Load RHSM configuration from file
    rhsm = Rhsm(module)

    state = module.params["state"]
    username = module.params["username"]
    password = module.params["password"]
    token = module.params["token"]
    server_hostname = module.params["server_hostname"]
    # TODO - no longer used? module.params["server_insecure"]
    # TODO - no longer used? module.params["server_prefix"]
    # TODO - no longer used? module.params["server_port"]
    # TODO - no longer used? module.params["rhsm_baseurl"]
    # TODO - no longer used? module.params["rhsm_repo_ca_cert"]
    auto_attach = module.params["auto_attach"]
    activationkey = module.params["activationkey"]
    org_id = module.params["org_id"]
    if activationkey and not org_id:
        module.fail_json(msg="org_id is required when using activationkey")
    environment = module.params["environment"]
    pool_ids = {}
    for value in module.params["pool_ids"]:
        if isinstance(value, dict):
            if len(value) != 1:
                module.fail_json(msg="Unable to parse pool_ids option.")
            pool_id, quantity = list(value.items())[0]
        else:
            pool_id, quantity = value, None
        pool_ids[pool_id] = quantity
    consumer_type = module.params["consumer_type"]
    consumer_name = module.params["consumer_name"]
    consumer_id = module.params["consumer_id"]
    force_register = module.params["force_register"]
    # TODO - no longer used? module.params["server_proxy_hostname"]
    # TODO - no longer used? module.params["server_proxy_port"]
    # TODO - no longer used? module.params["server_proxy_user"]
    # TODO - no longer used? module.params["server_proxy_password"]
    release = module.params["release"]
    syspurpose = module.params["syspurpose"]

    global SUBMAN_CMD
    SUBMAN_CMD = module.get_bin_path("subscription-manager", True)

    syspurpose_changed = False
    if syspurpose is not None:
        try:
            syspurpose_changed = SysPurpose().update_syspurpose(syspurpose)
        except Exception as err:
            module.fail_json(msg=f"Failed to update syspurpose attributes: {err}")

    # Ensure system is registered
    if state == "present":
        # Cache the status of the system before the changes
        was_registered = rhsm.is_registered

        # Register system
        if was_registered and not force_register:
            if syspurpose and "sync" in syspurpose and syspurpose["sync"] is True:
                try:
                    rhsm.sync_syspurpose()
                except Exception as e:
                    module.fail_json(msg=f"Failed to synchronize syspurpose attributes: {e}")
            if pool_ids:
                try:
                    result = rhsm.update_subscriptions_by_pool_ids(pool_ids)
                except Exception as e:
                    module.fail_json(msg=f"Failed to update subscriptions for '{server_hostname}': {e}")
                else:
                    module.exit_json(**result)
            else:
                if syspurpose_changed is True:
                    module.exit_json(changed=True, msg="Syspurpose attributes changed.")
                else:
                    module.exit_json(changed=False, msg="System already registered.")
        else:
            if not username and not activationkey and not token:
                module.fail_json(
                    msg="state is present but any of the following are missing: username, activationkey, token"
                )
            try:
                rhsm.enable()
                rhsm.configure(**module.params)
                rhsm.register(
                    was_registered,
                    username,
                    password,
                    token,
                    auto_attach,
                    activationkey,
                    org_id,
                    consumer_type,
                    consumer_name,
                    consumer_id,
                    force_register,
                    environment,
                    release,
                )
                if syspurpose and "sync" in syspurpose and syspurpose["sync"] is True:
                    rhsm.sync_syspurpose()
                if pool_ids:
                    subscribed_pool_ids = rhsm.subscribe_by_pool_ids(pool_ids)
                else:
                    subscribed_pool_ids = []
            except Exception as e:
                module.fail_json(msg=f"Failed to register with '{server_hostname}': {e}")
            else:
                module.exit_json(
                    changed=True,
                    msg=f"System successfully registered to '{server_hostname}'.",
                    subscribed_pool_ids=subscribed_pool_ids,
                )

    # Ensure system is *not* registered
    if state == "absent":
        if not rhsm.is_registered:
            module.exit_json(changed=False, msg="System already unregistered.")
        else:
            try:
                rhsm.unregister()
            except Exception as e:
                module.fail_json(msg=f"Failed to unregister: {e}")
            else:
                module.exit_json(changed=True, msg=f"System successfully unregistered from {server_hostname}.")


if __name__ == "__main__":
    main()



#!/usr/bin/python

# Copyright (c) 2015, Brian Coca <bcoca@ansible.com>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: runit
author:
  - James Sumners (@jsumners)
short_description: Manage runit services
description:
  - Controls runit services on remote hosts using the sv utility.
extends_documentation_fragment:
  - community.general._attributes
attributes:
  check_mode:
    support: full
  diff_mode:
    support: none
options:
  name:
    description:
      - Name of the service to manage.
    type: str
    required: true
  state:
    description:
      - V(started)/V(stopped) are idempotent actions that do not run commands unless necessary.
      - V(restarted) always bounces the service (sv restart) and V(killed) always bounces the service (sv force-stop).
      - V(reloaded) always sends a HUP (sv reload).
      - V(once) runs a normally downed sv once (sv once), not really an idempotent operation.
    type: str
    choices: [killed, once, reloaded, restarted, started, stopped]
  enabled:
    description:
      - Whether the service is enabled or not, if disabled it also implies stopped.
    type: bool
  service_dir:
    description:
      - Directory runsv watches for services.
    type: str
    default: /var/service
  service_src:
    description:
      - Directory where services are defined, the source of symlinks to O(service_dir).
    type: str
    default: /etc/sv
"""

EXAMPLES = r"""
- name: Start sv dnscache, if not running
  community.general.runit:
    name: dnscache
    state: started

- name: Stop sv dnscache, if running
  community.general.runit:
    name: dnscache
    state: stopped

- name: Kill sv dnscache, in all cases
  community.general.runit:
    name: dnscache
    state: killed

- name: Restart sv dnscache, in all cases
  community.general.runit:
    name: dnscache
    state: restarted

- name: Reload sv dnscache, in all cases
  community.general.runit:
    name: dnscache
    state: reloaded

- name: Use alternative sv directory location
  community.general.runit:
    name: dnscache
    state: reloaded
    service_dir: /run/service
"""

import os
import re

from ansible.module_utils.basic import AnsibleModule


class Sv:
    """
    Main class that handles daemontools, can be subclassed and overridden in case
    we want to use a 'derivative' like encore, s6, etc
    """

    def __init__(self, module):
        self.extra_paths = []
        self.report_vars = ["state", "enabled", "svc_full", "src_full", "pid", "duration", "full_state"]

        self.module = module

        self.name = module.params["name"]
        self.service_dir = module.params["service_dir"]
        self.service_src = module.params["service_src"]
        self.enabled = None
        self.full_state = None
        self.state = None
        self.pid = None
        self.duration = None

        self.svc_cmd = module.get_bin_path("sv", opt_dirs=self.extra_paths, required=True)
        self.svstat_cmd = module.get_bin_path("sv", opt_dirs=self.extra_paths)
        self.svc_full = f"{self.service_dir}/{self.name}"
        self.src_full = f"{self.service_src}/{self.name}"

        self.enabled = os.path.lexists(self.svc_full)
        if self.enabled:
            self.get_status()
        else:
            self.state = "stopped"

    def enable(self):
        if os.path.exists(self.src_full):
            try:
                os.symlink(self.src_full, self.svc_full)
            except OSError as e:
                self.module.fail_json(path=self.src_full, msg=f"Error while linking: {e}")
        else:
            self.module.fail_json(msg=f"Could not find source for service to enable ({self.src_full}).")

    def disable(self):
        self.execute_command([self.svc_cmd, "force-stop", self.src_full])
        try:
            os.unlink(self.svc_full)
        except OSError as e:
            self.module.fail_json(path=self.svc_full, msg=f"Error while unlinking: {e}")

    def get_status(self):
        (rc, out, err) = self.execute_command([self.svstat_cmd, "status", self.svc_full])

        if err is not None and err:
            self.full_state = self.state = err
        else:
            self.full_state = out
            # full_state *may* contain information about the logger:
            # "down: /etc/service/service-without-logger: 1s, normally up\n"
            # "down: /etc/service/updater: 127s, normally up; run: log: (pid 364) 263439s\n"
            full_state_no_logger = self.full_state.split("; ")[0]

            m = re.search(r"\(pid (\d+)\)", full_state_no_logger)
            if m:
                self.pid = m.group(1)

            m = re.search(r" (\d+)s", full_state_no_logger)
            if m:
                self.duration = m.group(1)

            if re.search(r"^run:", full_state_no_logger):
                self.state = "started"
            elif re.search(r"^down:", full_state_no_logger):
                self.state = "stopped"
            else:
                self.state = "unknown"
                return

    def started(self):
        return self.start()

    def start(self):
        return self.execute_command([self.svc_cmd, "start", self.svc_full])

    def stopped(self):
        return self.stop()

    def stop(self):
        return self.execute_command([self.svc_cmd, "stop", self.svc_full])

    def once(self):
        return self.execute_command([self.svc_cmd, "once", self.svc_full])

    def reloaded(self):
        return self.reload()

    def reload(self):
        return self.execute_command([self.svc_cmd, "reload", self.svc_full])

    def restarted(self):
        return self.restart()

    def restart(self):
        return self.execute_command([self.svc_cmd, "restart", self.svc_full])

    def killed(self):
        return self.kill()

    def kill(self):
        return self.execute_command([self.svc_cmd, "force-stop", self.svc_full])

    def execute_command(self, cmd):
        try:
            (rc, out, err) = self.module.run_command(cmd)
        except Exception as e:
            self.module.fail_json(msg=f"failed to execute: {e}")
        return rc, out, err

    def report(self):
        self.get_status()
        states = {}
        for k in self.report_vars:
            states[k] = self.__dict__[k]
        return states


def main():
    module = AnsibleModule(
        argument_spec=dict(
            name=dict(type="str", required=True),
            state=dict(type="str", choices=["killed", "once", "reloaded", "restarted", "started", "stopped"]),
            enabled=dict(type="bool"),
            service_dir=dict(type="str", default="/var/service"),
            service_src=dict(type="str", default="/etc/sv"),
        ),
        supports_check_mode=True,
    )

    module.run_command_environ_update = dict(LANGUAGE="C", LC_ALL="C")

    state = module.params["state"]
    enabled = module.params["enabled"]

    sv = Sv(module)
    changed = False

    if enabled is not None and enabled != sv.enabled:
        changed = True
        if not module.check_mode:
            try:
                if enabled:
                    sv.enable()
                else:
                    sv.disable()
            except OSError as e:
                module.fail_json(msg=f"Could not change service link: {e}")

    if state is not None and state != sv.state:
        changed = True
        if not module.check_mode:
            getattr(sv, state)()

    module.exit_json(changed=changed, sv=sv.report())


if __name__ == "__main__":
    main()



#!/usr/bin/python
#
# Scaleway VPC management module
#
# Copyright (c) Ansible project
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: scaleway_compute_private_network
short_description: Scaleway compute - private network management
version_added: 5.2.0
author: Pascal MANGIN (@pastral)
description:
  - This module add or remove a private network to a compute instance (U(https://developer.scaleway.com)).
extends_documentation_fragment:
  - community.general._scaleway
  - community.general._attributes
  - community.general._scaleway.actiongroup_scaleway

attributes:
  check_mode:
    support: full
  diff_mode:
    support: none
  action_group:
    version_added: 11.3.0

options:
  state:
    type: str
    description:
      - Indicate desired state of the VPC.
    default: present
    choices:
      - present
      - absent

  project:
    type: str
    description:
      - Project identifier.
    required: true

  region:
    type: str
    description:
      - Scaleway region to use (for example V(par1)).
    required: true
    choices:
      - ams1
      - EMEA-NL-EVS
      - ams2
      - ams3
      - par1
      - EMEA-FR-PAR1
      - par2
      - EMEA-FR-PAR2
      - par3
      - waw1
      - EMEA-PL-WAW1
      - waw2
      - waw3

  compute_id:
    type: str
    description:
      - ID of the compute instance (see M(community.general.scaleway_compute)).
    required: true

  private_network_id:
    type: str
    description:
      - ID of the private network (see M(community.general.scaleway_private_network)).
    required: true
"""

EXAMPLES = r"""
- name: Plug a VM to a private network
  community.general.scaleway_compute_private_network:
    project: '{{ scw_project }}'
    state: present
    region: par1
    compute_id: "12345678-f1e6-40ec-83e5-12345d67ed89"
    private_network_id: "22345678-f1e6-40ec-83e5-12345d67ed89"
  register: nicsvpc_creation_task

- name: Unplug a VM from a private network
  community.general.scaleway_compute_private_network:
    project: '{{ scw_project }}'
    state: absent
    region: par1
    compute_id: "12345678-f1e6-40ec-83e5-12345d67ed89"
    private_network_id: "22345678-f1e6-40ec-83e5-12345d67ed89"
"""

RETURN = r"""
scaleway_compute_private_network:
  description: Information on the VPC.
  returned: success when O(state=present)
  type: dict
  sample:
    {
      "created_at": "2022-01-15T11:11:12.676445Z",
      "id": "12345678-f1e6-40ec-83e5-12345d67ed89",
      "name": "network",
      "organization_id": "a123b4cd-ef5g-678h-90i1-jk2345678l90",
      "project_id": "a123b4cd-ef5g-678h-90i1-jk2345678l90",
      "tags": [
        "tag1",
        "tag2",
        "tag3",
        "tag4",
        "tag5"
      ],
      "updated_at": "2022-01-15T11:12:04.624837Z",
      "zone": "fr-par-2"
    }
"""

from ansible.module_utils.basic import AnsibleModule

from ansible_collections.community.general.plugins.module_utils._scaleway import (
    SCALEWAY_LOCATION,
    Scaleway,
    scaleway_argument_spec,
)


def get_nics_info(api, compute_id, private_network_id):
    response = api.get(f"servers/{compute_id}/private_nics")
    if not response.ok:
        msg = f"Error during get servers information: {response.info['msg']}: '{response.json['message']}' ({response.json})"
        api.module.fail_json(msg=msg)

    i = 0
    list_nics = response.json["private_nics"]

    while i < len(list_nics):
        if list_nics[i]["private_network_id"] == private_network_id:
            return list_nics[i]
        i += 1

    return None


def present_strategy(api, compute_id, private_network_id):
    changed = False
    nic = get_nics_info(api, compute_id, private_network_id)
    if nic is not None:
        return changed, nic

    data = {"private_network_id": private_network_id}
    changed = True
    if api.module.check_mode:
        return changed, {"status": "a private network would be add to a server"}

    response = api.post(path=f"servers/{compute_id}/private_nics", data=data)

    if not response.ok:
        api.module.fail_json(
            msg=f"Error when adding a private network to a server [{response.status_code}: {response.json}]"
        )

    return changed, response.json


def absent_strategy(api, compute_id, private_network_id):
    changed = False
    nic = get_nics_info(api, compute_id, private_network_id)
    if nic is None:
        return changed, {}

    changed = True
    if api.module.check_mode:
        return changed, {"status": "private network would be destroyed"}

    response = api.delete(f"servers/{compute_id}/private_nics/{nic['id']}")

    if not response.ok:
        api.module.fail_json(
            msg=f"Error deleting private network from server [{response.status_code}: {response.json}]"
        )

    return changed, response.json


def core(module):
    compute_id = module.params["compute_id"]
    pn_id = module.params["private_network_id"]

    region = module.params["region"]
    module.params["api_url"] = SCALEWAY_LOCATION[region]["api_endpoint"]

    api = Scaleway(module=module)
    if module.params["state"] == "absent":
        changed, summary = absent_strategy(api=api, compute_id=compute_id, private_network_id=pn_id)
    else:
        changed, summary = present_strategy(api=api, compute_id=compute_id, private_network_id=pn_id)
    module.exit_json(changed=changed, scaleway_compute_private_network=summary)


def main():
    argument_spec = scaleway_argument_spec()
    argument_spec.update(
        dict(
            state=dict(default="present", choices=["absent", "present"]),
            project=dict(required=True),
            region=dict(required=True, choices=list(SCALEWAY_LOCATION.keys())),
            compute_id=dict(required=True),
            private_network_id=dict(required=True),
        )
    )
    module = AnsibleModule(
        argument_spec=argument_spec,
        supports_check_mode=True,
    )

    core(module)


if __name__ == "__main__":
    main()



#!/usr/bin/python
# Copyright (c) 2017, Joseph Benden <joe@benden.us>
#
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: xfconf
author:
  - "Joseph Benden (@jbenden)"
  - "Alexei Znamensky (@russoz)"
short_description: Edit XFCE4 Configurations
description:
  - This module allows for the manipulation of Xfce 4 Configuration with the help of C(xfconf-query).
seealso:
  - name: xfconf-query(1) man page
    description: Manual page of the C(xfconf-query) tool at the XFCE documentation site.
    link: 'https://docs.xfce.org/xfce/xfconf/xfconf-query'

  - name: xfconf - Configuration Storage System
    description: XFCE documentation for the Xfconf configuration system.
    link: 'https://docs.xfce.org/xfce/xfconf/start'

  - name: xfce4-settings-editor - Settings Editor
    description: XFCE documentation for the graphical editor for configuration settings.
    link: https://docs.xfce.org/xfce/xfce4-settings/editor#change_properties

extends_documentation_fragment:
  - community.general._attributes

attributes:
  check_mode:
    support: full
  diff_mode:
    support: full

options:
  channel:
    description:
      - A Xfconf preference channel is a top-level tree key, inside of the Xfconf repository that corresponds to the location
        for which all application properties/keys are stored. See man xfconf-query(1).
    required: true
    type: str
  property:
    description:
      - A Xfce preference key is an element in the Xfconf repository that corresponds to an application preference. See man
        xfconf-query(1).
    required: true
    type: str
  value:
    description:
      - Preference properties typically have simple values such as strings, integers, or lists of strings and integers. See
        man xfconf-query(1).
    type: list
    elements: raw
  value_type:
    description:
      - The type of value being set.
      - When providing more than one O(value_type), the length of the list must be equal to the length of O(value).
      - If only one O(value_type) is provided, but O(value) contains more than on element, that O(value_type) is applied to
        all elements of O(value).
      - If the O(property) being set is an array and it can possibly have only one element in the array, then O(force_array=true)
        must be used to ensure that C(xfconf-query) interprets the value as an array rather than a scalar.
      - Support for V(uchar), V(char), V(uint64), and V(int64) has been added in community.general 4.8.0.
    type: list
    elements: str
    choices: [string, int, double, bool, uint, uchar, char, uint64, int64, float]
  state:
    type: str
    description:
      - The action to take upon the property/value.
      - The state V(get) has been removed in community.general 5.0.0. Please use the module M(community.general.xfconf_info)
        instead.
      - Xfce4 may, and usually does, have default values that come with the system packages.
        You can set new values for these default properties and override the their values.
        However, whey you use O(state=absent), the module executes the command C(xfconf-query reset) for the specified property and
        that only removes user-configured properties, so those properties are B(not removed), but rather they go back to the default values.
    choices: [present, absent]
    default: "present"
  force_array:
    description:
      - Force array even if only one element.
    type: bool
    default: false
    aliases: ['array']
    version_added: 1.0.0
"""

EXAMPLES = r"""
- name: Change the DPI to "192"
  xfconf:
    channel: "xsettings"
    property: "/Xft/DPI"
    value_type: "int"
    value: "192"

- name: Set workspace names (4)
  xfconf:
    channel: xfwm4
    property: /general/workspace_names
    value_type: string
    value: ['Main', 'Work1', 'Work2', 'Tmp']

- name: Set workspace names (1)
  xfconf:
    channel: xfwm4
    property: /general/workspace_names
    value_type: string
    value: ['Main']
    force_array: true
"""

RETURN = r"""
channel:
  description: The channel specified in the module parameters.
  returned: success
  type: str
  sample: "xsettings"
property:
  description: The property specified in the module parameters.
  returned: success
  type: str
  sample: "/Xft/DPI"
value_type:
  description:
    - The type of the value that was changed (V(none) for O(state=reset)). Either a single string value or a list of strings
      for array types.
    - This is a string or a list of strings.
  returned: success
  type: any
  sample: '"int" or ["str", "str", "str"]'
value:
  description:
    - The value of the preference key after executing the module. Either a single string value or a list of strings for array
      types.
    - This is a string or a list of strings.
  returned: success
  type: any
  sample: "'192' or ['orange', 'yellow', 'violet']"
previous_value:
  description:
    - The value of the preference key before executing the module. Either a single string value or a list of strings for array
      types.
    - This is a string or a list of strings.
  returned: success
  type: any
  sample: '"96" or ["red", "blue", "green"]'
cmd:
  description:
    - A list with the resulting C(xfconf-query) command executed by the module.
  returned: success
  type: list
  elements: str
  version_added: 5.4.0
  sample:
    - /usr/bin/xfconf-query
    - --channel
    - xfce4-panel
    - --property
    - /plugins/plugin-19/timezone
    - --create
    - --type
    - string
    - --set
    - Pacific/Auckland
version:
  description:
    - The version of the C(xfconf-query) command.
  returned: success
  type: str
  sample: 4.18.1
  version_added: 10.2.0
"""

from ansible.module_utils.parsing.convert_bool import boolean

from ansible_collections.community.general.plugins.module_utils._module_helper import StateModuleHelper
from ansible_collections.community.general.plugins.module_utils._xfconf import get_xfconf_version, xfconf_runner


class XFConfProperty(StateModuleHelper):
    change_params = ("value",)
    diff_params = ("value",)
    output_params = ("property", "channel", "value")
    module = dict(
        argument_spec=dict(
            state=dict(type="str", choices=("present", "absent"), default="present"),
            channel=dict(type="str", required=True),
            property=dict(type="str", required=True),
            value_type=dict(
                type="list",
                elements="str",
                choices=("string", "int", "double", "bool", "uint", "uchar", "char", "uint64", "int64", "float"),
            ),
            value=dict(type="list", elements="raw"),
            force_array=dict(type="bool", default=False, aliases=["array"]),
        ),
        required_if=[("state", "present", ["value", "value_type"])],
        required_together=[("value", "value_type")],
        supports_check_mode=True,
    )

    def __init_module__(self):
        self.runner = xfconf_runner(self.module)
        self.vars.version = get_xfconf_version(self.runner)
        self.does_not = f'Property "{self.vars.property}" does not exist on channel "{self.vars.channel}".'
        self.vars.set("previous_value", self._get())
        self.vars.set("type", self.vars.value_type)
        self.vars.set_meta("value", initial_value=self.vars.previous_value)

    def process_command_output(self, rc, out, err):
        if err.rstrip() == self.does_not:
            return None
        if rc or len(err):
            self.do_raise(f"xfconf-query failed with error (rc={rc}): {err}")

        result = out.rstrip()
        if "Value is an array with" in result:
            result = result.split("\n")
            if len(result) > 1:
                result.pop(0)
                result.pop(0)
            else:
                return []

        return result

    def _get(self):
        with self.runner("channel property", output_process=self.process_command_output) as ctx:
            return ctx.run()

    def state_absent(self):
        with self.runner("channel property reset", check_mode_skip=True) as ctx:
            ctx.run(reset=True)
            self.vars.stdout = ctx.results_out
            self.vars.stderr = ctx.results_err
            self.vars.cmd = ctx.cmd
            self.vars.set("run_info", ctx.run_info, verbosity=4)
        self.vars.value = None

    def state_present(self):
        value_type = self.vars.value_type

        values_len = len(self.vars.value)
        types_len = len(value_type)

        if types_len == 1:
            # use one single type for the entire list
            value_type = value_type * values_len
        elif types_len != values_len:
            # or complain if lists' lengths are different
            self.do_raise('Number of elements in "value" and "value_type" must be the same')

        # stringify all values - in the CLI they will all be happy strings anyway
        # and by doing this here the rest of the code can be agnostic to it
        # bool values are normalized to 'true'/'false' to match xfconf-query output format
        self.vars.value = [
            ("true" if boolean(v) else "false") if vt == "bool" else str(v)
            for v, vt in zip(self.vars.value, value_type)
        ]

        # calculates if it is an array
        self.vars.is_array = bool(self.vars.force_array) or isinstance(self.vars.previous_value, list) or values_len > 1

        with self.runner("channel property create force_array values_and_types", check_mode_skip=True) as ctx:
            ctx.run(create=True, force_array=self.vars.is_array, values_and_types=(self.vars.value, value_type))
            self.vars.stdout = ctx.results_out
            self.vars.stderr = ctx.results_err
            self.vars.cmd = ctx.cmd
            self.vars.set("run_info", ctx.run_info, verbosity=4)

        if not self.vars.is_array:
            self.vars.value = self.vars.value[0]
            self.vars.type = value_type[0]
        else:
            self.vars.type = value_type


def main():
    XFConfProperty.execute()


if __name__ == "__main__":
    main()



#!/usr/bin/python
# Copyright Ansible Project
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: uptimerobot
short_description: Pause and start Uptime Robot monitoring
deprecated:
  removed_in: 15.0.0
  why: The module uses the Uptime Robot API v1, which was retired and is no longer available.
  alternative: Use the Uptime Robot API v3 directly with the M(ansible.builtin.uri) module.
description:
  - This module lets you start and pause Uptime Robot Monitoring.
author: "Nate Kingsley (@nate-kingsley)"
requirements:
  - Valid Uptime Robot API Key
extends_documentation_fragment:
  - community.general._attributes
attributes:
  check_mode:
    support: none
  diff_mode:
    support: none
options:
  state:
    type: str
    description:
      - Define whether or not the monitor should be running or paused.
    required: true
    choices: ["started", "paused"]
  monitorid:
    type: str
    description:
      - ID of the monitor to check.
    required: true
  apikey:
    type: str
    description:
      - Uptime Robot API key.
    required: true
notes:
  - Support for adding and removing monitors and alert contacts has not yet been implemented.
"""

EXAMPLES = r"""
- name: Pause the monitor with an ID of 12345
  community.general.uptimerobot:
    monitorid: 12345
    apikey: 12345-1234512345
    state: paused

- name: Start the monitor with an ID of 12345
  community.general.uptimerobot:
    monitorid: 12345
    apikey: 12345-1234512345
    state: started
"""

import json
from urllib.parse import urlencode

from ansible.module_utils.basic import AnsibleModule
from ansible.module_utils.common.text.converters import to_text
from ansible.module_utils.urls import fetch_url

API_BASE = "https://api.uptimerobot.com/"

API_ACTIONS = dict(status="getMonitors?", editMonitor="editMonitor?")

API_FORMAT = "json"
API_NOJSONCALLBACK = 1
CHANGED_STATE = False
SUPPORTS_CHECK_MODE = False


def checkID(module, params):
    data = urlencode(params)
    full_uri = API_BASE + API_ACTIONS["status"] + data
    req, info = fetch_url(module, full_uri)
    result = to_text(req.read())
    jsonresult = json.loads(result)
    req.close()
    return jsonresult


def startMonitor(module, params):
    params["monitorStatus"] = 1
    data = urlencode(params)
    full_uri = API_BASE + API_ACTIONS["editMonitor"] + data
    req, info = fetch_url(module, full_uri)
    result = to_text(req.read())
    jsonresult = json.loads(result)
    req.close()
    return jsonresult["stat"]


def pauseMonitor(module, params):
    params["monitorStatus"] = 0
    data = urlencode(params)
    full_uri = API_BASE + API_ACTIONS["editMonitor"] + data
    req, info = fetch_url(module, full_uri)
    result = to_text(req.read())
    jsonresult = json.loads(result)
    req.close()
    return jsonresult["stat"]


def main():
    module = AnsibleModule(
        argument_spec=dict(
            state=dict(required=True, choices=["started", "paused"]),
            apikey=dict(required=True, no_log=True),
            monitorid=dict(required=True),
        ),
        supports_check_mode=SUPPORTS_CHECK_MODE,
    )

    params = dict(
        apiKey=module.params["apikey"],
        monitors=module.params["monitorid"],
        monitorID=module.params["monitorid"],
        format=API_FORMAT,
        noJsonCallback=API_NOJSONCALLBACK,
    )

    check_result = checkID(module, params)

    if check_result["stat"] != "ok":
        module.fail_json(msg="failed", result=check_result["message"])

    if module.params["state"] == "started":
        monitor_result = startMonitor(module, params)
    else:
        monitor_result = pauseMonitor(module, params)

    module.exit_json(msg="success", result=monitor_result)


if __name__ == "__main__":
    main()



#!/usr/bin/python
# Copyright (c) 2018, Milan Ilic <milani@nordeus.com>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

# Make coding more python3-ish
from __future__ import annotations

DOCUMENTATION = r"""
module: one_image
short_description: Manages OpenNebula images
description:
  - Manages OpenNebula images.
requirements:
  - pyone
extends_documentation_fragment:
  - community.general._opennebula
  - community.general._attributes
attributes:
  check_mode:
    support: full
  diff_mode:
    support: none
options:
  id:
    description:
      - A O(id) of the image you would like to manage.
    type: int
  name:
    description:
      - A O(name) of the image you would like to manage.
      - Required if O(create=true).
    type: str
  state:
    description:
      - V(present) - state that is used to manage the image.
      - V(absent) - delete the image.
      - V(cloned) - clone the image.
      - V(renamed) - rename the image to the O(new_name).
    choices: ["present", "absent", "cloned", "renamed"]
    default: present
    type: str
  enabled:
    description:
      - Whether the image should be enabled or disabled.
    type: bool
  new_name:
    description:
      - A name that is assigned to the existing or new image.
      - In the case of cloning, by default O(new_name) is set to the name of the origin image with the prefix 'Copy of'.
    type: str
  persistent:
    description:
      - Whether the image should be persistent or non-persistent.
    type: bool
    version_added: 9.5.0
  create:
    description:
      - Whether the image should be created if not present.
      - This is ignored if O(state=absent).
    type: bool
    version_added: 10.0.0
  template:
    description:
      - Use with O(create=true) to specify image template.
    type: str
    version_added: 10.0.0
  datastore_id:
    description:
      - Use with O(create=true) to specify datastore for image.
    type: int
    version_added: 10.0.0
  wait_timeout:
    description:
      - Seconds to wait until image is ready, deleted or cloned.
    type: int
    default: 60
    version_added: 10.0.0
author:
  - "Milan Ilic (@ilicmilan)"
"""

EXAMPLES = r"""
- name: Fetch the IMAGE by id
  community.general.one_image:
    id: 45
  register: result

- name: Print the IMAGE properties
  ansible.builtin.debug:
    var: result

- name: Rename existing IMAGE
  community.general.one_image:
    id: 34
    state: renamed
    new_name: bar-image

- name: Disable the IMAGE by id
  community.general.one_image:
    id: 37
    enabled: false

- name: Make the IMAGE persistent
  community.general.one_image:
    id: 37
    persistent: true

- name: Enable the IMAGE by name
  community.general.one_image:
    name: bar-image
    enabled: true

- name: Clone the IMAGE by name
  community.general.one_image:
    name: bar-image
    state: cloned
    new_name: bar-image-clone
  register: result

- name: Delete the IMAGE by id
  community.general.one_image:
    id: '{{ result.id }}'
    state: absent

- name: Make sure IMAGE is present
  community.general.one_image:
    name: myyy-image
    state: present
    create: true
    datastore_id: 100
    template: |
      PATH = "/var/tmp/image"
      TYPE = "OS"
      SIZE = 20512
      FORMAT = "qcow2"
      PERSISTENT = "Yes"
      DEV_PREFIX = "vd"

- name: Make sure IMAGE is present with a longer timeout
  community.general.one_image:
    name: big-image
    state: present
    create: true
    datastore_id: 100
    wait_timeout: 900
    template: |-
      PATH = "https://192.0.2.200/repo/tipa_image.raw"
      TYPE = "OS"
      SIZE = 82048
      FORMAT = "raw"
      PERSISTENT = "Yes"
      DEV_PREFIX = "vd"
"""

RETURN = r"""
id:
  description: Image ID.
  type: int
  returned: when O(state=present), O(state=cloned), or O(state=renamed)
  sample: 153
name:
  description: Image name.
  type: str
  returned: when O(state=present), O(state=cloned), or O(state=renamed)
  sample: app1
group_id:
  description: Image's group ID.
  type: int
  returned: when O(state=present), O(state=cloned), or O(state=renamed)
  sample: 1
group_name:
  description: Image's group name.
  type: str
  returned: when O(state=present), O(state=cloned), or O(state=renamed)
  sample: one-users
owner_id:
  description: Image's owner ID.
  type: int
  returned: when O(state=present), O(state=cloned), or O(state=renamed)
  sample: 143
owner_name:
  description: Image's owner name.
  type: str
  returned: when O(state=present), O(state=cloned), or O(state=renamed)
  sample: ansible-test
state:
  description: State of image instance.
  type: str
  returned: when O(state=present), O(state=cloned), or O(state=renamed)
  sample: READY
used:
  description: Is image in use.
  type: bool
  returned: when O(state=present), O(state=cloned), or O(state=renamed)
  sample: true
running_vms:
  description: Count of running vms that use this image.
  type: int
  returned: when O(state=present), O(state=cloned), or O(state=renamed)
  sample: 7
permissions:
  description: The image's permissions.
  type: dict
  returned: when O(state=present), O(state=cloned), or O(state=renamed)
  version_added: 9.5.0
  contains:
    owner_u:
      description: The image's owner USAGE permissions.
      type: str
      sample: 1
    owner_m:
      description: The image's owner MANAGE permissions.
      type: str
      sample: 0
    owner_a:
      description: The image's owner ADMIN permissions.
      type: str
      sample: 0
    group_u:
      description: The image's group USAGE permissions.
      type: str
      sample: 0
    group_m:
      description: The image's group MANAGE permissions.
      type: str
      sample: 0
    group_a:
      description: The image's group ADMIN permissions.
      type: str
      sample: 0
    other_u:
      description: The image's other users USAGE permissions.
      type: str
      sample: 0
    other_m:
      description: The image's other users MANAGE permissions.
      type: str
      sample: 0
    other_a:
      description: The image's other users ADMIN permissions.
      type: str
      sample: 0
  sample:
    owner_u: 1
    owner_m: 0
    owner_a: 0
    group_u: 0
    group_m: 0
    group_a: 0
    other_u: 0
    other_m: 0
    other_a: 0
type:
  description: The image's type.
  type: str
  sample: 0
  returned: when O(state=present), O(state=cloned), or O(state=renamed)
  version_added: 9.5.0
disk_type:
  description: The image's format type.
  type: str
  sample: 0
  returned: when O(state=present), O(state=cloned), or O(state=renamed)
  version_added: 9.5.0
persistent:
  description: The image's persistence status (1 means true, 0 means false).
  type: int
  sample: 1
  returned: when O(state=present), O(state=cloned), or O(state=renamed)
  version_added: 9.5.0
source:
  description: The image's source.
  type: str
  sample: /var/lib/one//datastores/100/somerandomstringxd
  returned: when O(state=present), O(state=cloned), or O(state=renamed)
path:
  description: The image's filesystem path.
  type: str
  sample: /var/tmp/hello.qcow2
  returned: when O(state=present), O(state=cloned), or O(state=renamed)
  version_added: 9.5.0
fstype:
  description: The image's filesystem type.
  type: str
  sample: ext4
  returned: when O(state=present), O(state=cloned), or O(state=renamed)
  version_added: 9.5.0
size:
  description: The image's size in MegaBytes.
  type: int
  sample: 10000
  returned: when O(state=present), O(state=cloned), or O(state=renamed)
  version_added: 9.5.0
cloning_ops:
  description: The image's cloning operations per second.
  type: int
  sample: 0
  returned: when O(state=present), O(state=cloned), or O(state=renamed)
  version_added: 9.5.0
cloning_id:
  description: The image's cloning ID.
  type: int
  sample: -1
  returned: when O(state=present), O(state=cloned), or O(state=renamed)
  version_added: 9.5.0
target_snapshot:
  description: The image's target snapshot.
  type: int
  sample: 1
  returned: when O(state=present), O(state=cloned), or O(state=renamed)
  version_added: 9.5.0
datastore_id:
  description: The image's datastore ID.
  type: int
  sample: 100
  returned: when O(state=present), O(state=cloned), or O(state=renamed)
  version_added: 9.5.0
datastore:
  description: The image's datastore name.
  type: int
  sample: image_datastore
  returned: when O(state=present), O(state=cloned), or O(state=renamed)
  version_added: 9.5.0
vms:
  description: The image's list of VM ID's.
  type: list
  elements: int
  returned: when O(state=present), O(state=cloned), or O(state=renamed)
  sample:
    - 1
    - 2
    - 3
  version_added: 9.5.0
clones:
  description: The image's list of clones ID's.
  type: list
  elements: int
  returned: when O(state=present), O(state=cloned), or O(state=renamed)
  sample:
    - 1
    - 2
    - 3
  version_added: 9.5.0
app_clones:
  description: The image's list of app_clones ID's.
  type: list
  elements: int
  returned: when O(state=present), O(state=cloned), or O(state=renamed)
  sample:
    - 1
    - 2
    - 3
  version_added: 9.5.0
snapshots:
  description: The image's list of snapshots.
  type: list
  returned: when O(state=present), O(state=cloned), or O(state=renamed)
  version_added: 9.5.0
  sample:
    - date: 123123
      parent: 1
      size: 10228
      allow_orphans: 1
      children: 0
      active: 1
      name: SampleName
"""


import time

from ansible_collections.community.general.plugins.module_utils._opennebula import OpenNebulaModule

IMAGE_STATES = [
    "INIT",
    "READY",
    "USED",
    "DISABLED",
    "LOCKED",
    "ERROR",
    "CLONE",
    "DELETE",
    "USED_PERS",
    "LOCKED_USED",
    "LOCKED_USED_PERS",
]


class ImageModule(OpenNebulaModule):
    def __init__(self):
        argument_spec = dict(
            id=dict(type="int"),
            name=dict(type="str"),
            state=dict(type="str", choices=["present", "absent", "cloned", "renamed"], default="present"),
            enabled=dict(type="bool"),
            new_name=dict(type="str"),
            persistent=dict(type="bool"),
            create=dict(type="bool"),
            template=dict(type="str"),
            datastore_id=dict(type="int"),
            wait_timeout=dict(type="int", default=60),
        )
        required_if = [
            ["state", "renamed", ["id"]],
            ["create", True, ["template", "datastore_id", "name"]],
        ]
        mutually_exclusive = [
            ["id", "name"],
        ]

        OpenNebulaModule.__init__(
            self,
            argument_spec,
            supports_check_mode=True,
            mutually_exclusive=mutually_exclusive,
            required_if=required_if,
        )

    def run(self, one, module, result):
        params = module.params
        id = params.get("id")
        name = params.get("name")
        desired_state = params.get("state")
        enabled = params.get("enabled")
        new_name = params.get("new_name")
        persistent = params.get("persistent")
        create = params.get("create")
        template = params.get("template")
        datastore_id = params.get("datastore_id")
        wait_timeout = params.get("wait_timeout")

        self.result = {}

        image = self.get_image_instance(id, name)
        if not image and desired_state != "absent":
            if create:
                self.result = self.create_image(name, template, datastore_id, wait_timeout)
            # Using 'if id:' doesn't work properly when id=0
            elif id is not None:
                module.fail_json(msg=f"There is no image with id={id}")
            elif name is not None:
                module.fail_json(msg=f"There is no image with name={name}")

        if desired_state == "absent":
            self.result = self.delete_image(image, wait_timeout)
        else:
            if persistent is not None:
                self.result = self.change_persistence(image, persistent)
            if enabled is not None:
                self.result = self.enable_image(image, enabled)
            if desired_state == "cloned":
                self.result = self.clone_image(image, new_name, wait_timeout)
            elif desired_state == "renamed":
                self.result = self.rename_image(image, new_name)

        self.exit()

    def get_image(self, predicate):
        # Filter -2 means fetch all images user can Use
        pool = self.one.imagepool.info(-2, -1, -1, -1)

        for image in pool.IMAGE:
            if predicate(image):
                return image

        return None

    def get_image_by_name(self, image_name):
        return self.get_image(lambda image: image_name == image.NAME)

    def get_image_by_id(self, image_id):
        return self.get_image(lambda image: image_id == image.ID)

    def get_image_instance(self, requested_id, requested_name):
        # Using 'if requested_id:' doesn't work properly when requested_id=0
        if requested_id is not None:
            return self.get_image_by_id(requested_id)
        else:
            return self.get_image_by_name(requested_name)

    def create_image(self, image_name, template, datastore_id, wait_timeout):
        if not self.module.check_mode:
            image_id = self.one.image.allocate(f'NAME = "{image_name}"\n{template}', datastore_id)
            self.wait_for_ready(image_id, wait_timeout)
            image = self.get_image_by_id(image_id)
            result = self.get_image_info(image)

        result["changed"] = True
        return result

    def wait_for_ready(self, image_id, wait_timeout=60):
        start_time = time.time()

        while (time.time() - start_time) < wait_timeout:
            image = self.one.image.info(image_id)
            state = image.STATE

            if state in [IMAGE_STATES.index("ERROR")]:
                self.module.fail_json(msg=f"Got an ERROR state: {image.TEMPLATE['ERROR']}")

            if state in [IMAGE_STATES.index("READY")]:
                return True

            time.sleep(1)
        self.module.fail_json(msg="Wait timeout has expired!")

    def wait_for_delete(self, image_id, wait_timeout=60):
        start_time = time.time()

        while (time.time() - start_time) < wait_timeout:
            # It might be already deleted by the time this function is called
            try:
                image = self.one.image.info(image_id)
            except Exception:
                check_image = self.get_image_instance(image_id)
                if not check_image:
                    return True

            state = image.STATE

            if state in [IMAGE_STATES.index("DELETE")]:
                return True

            time.sleep(1)

        self.module.fail_json(msg="Wait timeout has expired!")

    def enable_image(self, image, enable):
        image = self.one.image.info(image.ID)
        changed = False

        state = image.STATE

        if state not in [IMAGE_STATES.index("READY"), IMAGE_STATES.index("DISABLED"), IMAGE_STATES.index("ERROR")]:
            if enable:
                self.module.fail_json(msg=f"Cannot enable {IMAGE_STATES[state]} image!")
            else:
                self.module.fail_json(msg=f"Cannot disable {IMAGE_STATES[state]} image!")

        if (enable and state != IMAGE_STATES.index("READY")) or (
            not enable and state != IMAGE_STATES.index("DISABLED")
        ):
            changed = True

        if changed and not self.module.check_mode:
            self.one.image.enable(image.ID, enable)

        result = self.get_image_info(image)
        result["changed"] = changed

        return result

    def change_persistence(self, image, enable):
        image = self.one.image.info(image.ID)
        changed = False

        state = image.STATE

        if state not in [IMAGE_STATES.index("READY"), IMAGE_STATES.index("DISABLED"), IMAGE_STATES.index("ERROR")]:
            if enable:
                self.module.fail_json(msg=f"Cannot enable persistence for {IMAGE_STATES[state]} image!")
            else:
                self.module.fail_json(msg=f"Cannot disable persistence for {IMAGE_STATES[state]} image!")

        if (enable and state != IMAGE_STATES.index("READY")) or (
            not enable and state != IMAGE_STATES.index("DISABLED")
        ):
            changed = True

        if changed and not self.module.check_mode:
            self.one.image.persistent(image.ID, enable)

        result = self.get_image_info(image)
        result["changed"] = changed

        return result

    def clone_image(self, image, new_name, wait_timeout):
        if new_name is None:
            new_name = f"Copy of {image.NAME}"

        tmp_image = self.get_image_by_name(new_name)
        if tmp_image:
            result = self.get_image_info(image)
            result["changed"] = False
            return result

        if IMAGE_STATES.index("DISABLED") == image.STATE:
            self.module.fail_json(msg="Cannot clone DISABLED image")

        if not self.module.check_mode:
            new_id = self.one.image.clone(image.ID, new_name)
            self.wait_for_ready(new_id, wait_timeout)
            image = self.one.image.info(new_id)

        result = self.get_image_info(image)
        result["changed"] = True

        return result

    def rename_image(self, image, new_name):
        if new_name is None:
            self.module.fail_json(msg="'new_name' option has to be specified when the state is 'renamed'")

        if new_name == image.NAME:
            result = self.get_image_info(image)
            result["changed"] = False
            return result

        tmp_image = self.get_image_by_name(new_name)
        if tmp_image:
            self.module.fail_json(msg=f"Name '{new_name}' is already taken by IMAGE with id={tmp_image.ID!s}")

        if not self.module.check_mode:
            self.one.image.rename(image.ID, new_name)

        result = self.get_image_info(image)
        result["changed"] = True
        return result

    def delete_image(self, image, wait_timeout):
        if not image:
            return {"changed": False}

        if image.RUNNING_VMS > 0:
            self.module.fail_json(msg=f"Cannot delete image. There are {image.RUNNING_VMS!s} VMs using it.")

        if not self.module.check_mode:
            self.one.image.delete(image.ID)
            self.wait_for_delete(image.ID, wait_timeout)

        return {"changed": True}


def main():
    ImageModule().run_module()


if __name__ == "__main__":
    main()



#!/usr/bin/python

# Copyright (c) 2015, Linus Unnebäck <linus@folkdatorn.se>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: make
short_description: Run targets in a Makefile
requirements:
  - make
author: Linus Unnebäck (@LinusU) <linus@folkdatorn.se>
description:
  - Run targets in a Makefile.
extends_documentation_fragment:
  - community.general._attributes
attributes:
  check_mode:
    support: full
  diff_mode:
    support: none
options:
  chdir:
    description:
      - Change to this directory before running make.
    type: path
    required: true
  file:
    description:
      - Use a custom Makefile.
    type: path
  jobs:
    description:
      - Set the number of make jobs to run concurrently.
      - Typically if set, this would be the number of processors and/or threads available to the machine.
      - This is not supported by all make implementations.
    type: int
    version_added: 2.0.0
  make:
    description:
      - Use a specific make binary.
    type: path
    version_added: '0.2.0'
  params:
    description:
      - Any extra parameters to pass to make.
      - If the value is empty, only the key is used. For example, V(FOO:) produces V(FOO), not V(FOO=).
    type: dict
  target:
    description:
      - The target to run.
      - Typically this would be something like V(install), V(test), or V(all).
      - O(target) and O(targets) are mutually exclusive.
    type: str
  targets:
    description:
      - The list of targets to run.
      - Typically this would be something like V(install), V(test), or V(all).
      - O(target) and O(targets) are mutually exclusive.
    type: list
    elements: str
    version_added: 7.2.0
"""

EXAMPLES = r"""
- name: Build the default target
  community.general.make:
    chdir: /home/ubuntu/cool-project

- name: Run 'install' target as root
  community.general.make:
    chdir: /home/ubuntu/cool-project
    target: install
  become: true

- name: Build 'all' target with extra arguments
  community.general.make:
    chdir: /home/ubuntu/cool-project
    target: all
    params:
      NUM_THREADS: 4
      BACKEND: lapack

- name: Build 'all' target with a custom Makefile
  community.general.make:
    chdir: /home/ubuntu/cool-project
    target: all
    file: /some-project/Makefile

- name: build arm64 kernel on FreeBSD, with 16 parallel jobs
  community.general.make:
    chdir: /usr/src
    jobs: 16
    target: buildkernel
    params:
      # This adds -DWITH_FDT to the command line:
      -DWITH_FDT:
      # The following adds TARGET=arm64 TARGET_ARCH=aarch64 to the command line:
      TARGET: arm64
      TARGET_ARCH: aarch64
"""

RETURN = r"""
chdir:
  description:
    - The value of the module parameter O(chdir).
  type: str
  returned: success
command:
  description:
    - The command built and executed by the module.
  type: str
  returned: success
  version_added: 6.5.0
file:
  description:
    - The value of the module parameter O(file).
  type: str
  returned: success
jobs:
  description:
    - The value of the module parameter O(jobs).
  type: int
  returned: success
params:
  description:
    - The value of the module parameter O(params).
  type: dict
  returned: success
target:
  description:
    - The value of the module parameter O(target).
  type: str
  returned: success
targets:
  description:
    - The value of the module parameter O(targets).
  type: str
  returned: success
  version_added: 7.2.0
"""

from shlex import quote as shlex_quote

from ansible.module_utils.basic import AnsibleModule


def run_command(command, module, check_rc=True):
    """
    Run a command using the module, return
    the result code and std{err,out} content.

    :param command: list of command arguments
    :param module: Ansible make module instance
    :return: return code, stdout content, stderr content
    """
    rc, out, err = module.run_command(command, check_rc=check_rc, cwd=module.params["chdir"])
    return rc, sanitize_output(out), sanitize_output(err)


def sanitize_output(output):
    """
    Sanitize the output string before we
    pass it to module.fail_json. Defaults
    the string to empty if it is None, else
    strips trailing newlines.

    :param output: output to sanitize
    :return: sanitized output
    """
    if output is None:
        return ""
    else:
        return output.rstrip("\r\n")


def main():
    module = AnsibleModule(
        argument_spec=dict(
            target=dict(type="str"),
            targets=dict(type="list", elements="str"),
            params=dict(type="dict"),
            chdir=dict(type="path", required=True),
            file=dict(type="path"),
            make=dict(type="path"),
            jobs=dict(type="int"),
        ),
        mutually_exclusive=[("target", "targets")],
        supports_check_mode=True,
    )

    make_path = module.params["make"]
    if make_path is None:
        # Build up the invocation of `make` we are going to use
        # For non-Linux OSes, prefer gmake (GNU make) over make
        make_path = module.get_bin_path("gmake", required=False)
        if not make_path:
            # Fall back to system make
            make_path = module.get_bin_path("make", required=True)
    if module.params["params"] is not None:
        make_parameters = [k + (f"={v!s}" if v is not None else "") for k, v in module.params["params"].items()]
    else:
        make_parameters = []

    # build command:
    # handle any make specific arguments included in params
    base_command = [make_path]
    if module.params["jobs"] is not None:
        jobs = str(module.params["jobs"])
        base_command.extend(["-j", jobs])
    if module.params["file"] is not None:
        base_command.extend(["-f", module.params["file"]])

    # add make target
    if module.params["target"]:
        base_command.append(module.params["target"])
    elif module.params["targets"]:
        base_command.extend(module.params["targets"])

    # add makefile parameters
    base_command.extend(make_parameters)

    # Check if the target is already up to date
    rc, out, err = run_command(base_command + ["-q"], module, check_rc=False)
    if module.check_mode:
        # If we've been asked to do a dry run, we only need
        # to report whether or not the target is up to date
        changed = rc != 0
    else:
        if rc == 0:
            # The target is up to date, so we don't have to
            #  do anything
            changed = False
        else:
            # The target isn't up to date, so we need to run it
            rc, out, err = run_command(base_command, module, check_rc=True)
            changed = True

    # We don't report the return code, as if this module failed
    # we would be calling fail_json from run_command, so even if
    # we had a non-zero return code, we did not fail. However, if
    # we report a non-zero return code here, we will be marked as
    # failed regardless of what we signal using the failed= kwarg.
    module.exit_json(
        changed=changed,
        failed=False,
        stdout=out,
        stderr=err,
        target=module.params["target"],
        targets=module.params["targets"],
        params=module.params["params"],
        chdir=module.params["chdir"],
        file=module.params["file"],
        jobs=module.params["jobs"],
        command=" ".join([shlex_quote(part) for part in base_command]),
    )


if __name__ == "__main__":
    main()



#!/usr/bin/python
# Copyright (c) 2016-2017 Hewlett Packard Enterprise Development LP
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: oneview_san_manager_info
short_description: Retrieve information about one or more of the OneView SAN Managers
description:
  - Retrieve information about one or more of the SAN Managers from OneView.
requirements:
  - hpOneView >= 2.0.1
author:
  - Felipe Bulsoni (@fgbulsoni)
  - Thiago Miotto (@tmiotto)
  - Adriane Cardozo (@adriane-cardozo)
attributes:
  check_mode:
    version_added: 3.3.0
    # This was backported to 2.5.4 and 1.3.11 as well, since this was a bugfix
options:
  provider_display_name:
    description:
      - Provider Display Name.
    type: str
  params:
    description:
      - List of params to delimit, filter and sort the list of resources.
      - 'Params allowed:'
      - 'V(start): The first item to return, using 0-based indexing.'
      - 'V(count): The number of resources to return.'
      - 'V(query): A general query string to narrow the list of resources returned.'
      - 'V(sort): The sort order of the returned data set.'
    type: dict
extends_documentation_fragment:
  - community.general._oneview
  - community.general._attributes
  - community.general._attributes.info_module
"""

EXAMPLES = r"""
- name: Gather information about all SAN Managers
  community.general.oneview_san_manager_info:
    config: /etc/oneview/oneview_config.json
  delegate_to: localhost
  register: result

- name: Print fetched information about SAN Managers
  ansible.builtin.debug:
    msg: "{{ result.san_managers }}"

- name: Gather paginated, filtered and sorted information about SAN Managers
  community.general.oneview_san_manager_info:
    config: /etc/oneview/oneview_config.json
    params:
      start: 0
      count: 3
      sort: name:ascending
      query: isInternal eq false
  delegate_to: localhost
  register: result

- name: Print fetched information about paginated, filtered and sorted list of SAN Managers
  ansible.builtin.debug:
    msg: "{{ result.san_managers }}"

- name: Gather information about a SAN Manager by provider display name
  community.general.oneview_san_manager_info:
    config: /etc/oneview/oneview_config.json
    provider_display_name: Brocade Network Advisor
  delegate_to: localhost
  register: result

- name: Print fetched information about SAN Manager found by provider display name
  ansible.builtin.debug:
    msg: "{{ result.san_managers }}"
"""

RETURN = r"""
san_managers:
  description: Has all the OneView information about the SAN Managers.
  returned: Always, but can be null.
  type: dict
"""

from ansible_collections.community.general.plugins.module_utils._oneview import OneViewModuleBase


class SanManagerInfoModule(OneViewModuleBase):
    argument_spec = dict(provider_display_name=dict(type="str"), params=dict(type="dict"))

    def __init__(self):
        super().__init__(
            additional_arg_spec=self.argument_spec,
            supports_check_mode=True,
        )
        self.resource_client = self.oneview_client.san_managers

    def execute_module(self):
        if self.module.params.get("provider_display_name"):
            provider_display_name = self.module.params["provider_display_name"]
            san_manager = self.oneview_client.san_managers.get_by_provider_display_name(provider_display_name)
            if san_manager:
                resources = [san_manager]
            else:
                resources = []
        else:
            resources = self.oneview_client.san_managers.get_all(**self.facts_params)

        return dict(changed=False, san_managers=resources)


def main():
    SanManagerInfoModule().run()


if __name__ == "__main__":
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright (c) 2021, Cisco Systems
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

DOCUMENTATION = r"""
module: organizations_wireless_devices_wireless_controllers_by_device_info
short_description: Information module for organizations _wireless _devices _wireless
  _controllers _by _device
description:
  - Information module for Organizations Wireless Devices Wireless Controllers By Device Info.
  - Get all organizations _wireless _devices _wireless _controllers _by _device.
  - List of Catalyst access points information.
version_added: '1.0.0'
extends_documentation_fragment:
  - cisco.meraki.module_info
  - cisco.meraki.module_info_pagination
author: Francisco Munoz (@fmunoz)
options:
  headers:
    description: Additional headers.
    type: dict
  organizationId:
    description:
      - Information module for Organizations Wireless Devices Wireless Controllers By Device Info.
      - OrganizationId path parameter. Organization ID.
    type: str
  networkIds:
    description:
      - Information module for Organizations Wireless Devices Wireless Controllers By Device Info.
      - >
        NetworkIds query parameter. Optional parameter to filter access points by
        network ID. This filter uses multiple exact matches.
    elements: str
    type: list
  serials:
    description:
      - Information module for Organizations Wireless Devices Wireless Controllers By Device Info.
      - >
        Serials query parameter. Optional parameter to filter access points by its
        cloud ID. This filter uses multiple exact matches.
    elements: str
    type: list
  controllerSerials:
    description:
      - Information module for Organizations Wireless Devices Wireless Controllers By Device Info.
      - >
        ControllerSerials query parameter. Optional parameter to filter access points
        by its wireless LAN controller cloud ID. This filter uses multiple exact matches.
    elements: str
    type: list
  perPage:
    description:
      - Information module for Organizations Wireless Devices Wireless Controllers By Device Info.
      - PerPage query parameter. The number of entries per page returned. Acceptable
        range is 3 - 1000. Default is 100.
    type: int
  startingAfter:
    description:
      - Information module for Organizations Wireless Devices Wireless Controllers By Device Info.
      - >
        StartingAfter query parameter. A token used by the server to indicate the
        start of the page. Often this is a timestamp or an ID but it is not limited
        to those. This parameter should not be defined by client applications. The
        link for the first, last, prev, or next page in the HTTP Link header should
        define it.
    type: str
  endingBefore:
    description:
      - Information module for Organizations Wireless Devices Wireless Controllers By Device Info.
      - >
        EndingBefore query parameter. A token used by the server to indicate the end
        of the page. Often this is a timestamp or an ID but it is not limited to those.
        This parameter should not be defined by client applications. The link for
        the first, last, prev, or next page in the HTTP Link header should define
        it.
    type: str
requirements:
  - meraki >= 2.4.9
  - python >= 3.5
seealso:
  - name: Cisco Meraki documentation for wireless getOrganizationWirelessDevicesWirelessControllersByDevice
    description: Complete reference of the getOrganizationWirelessDevicesWirelessControllersByDevice
      API.
    link: https://developer.cisco.com/meraki/api-v1/#!get-organization-wireless-devices-wireless-controllers-by-device
notes:
  - SDK Method used are
    wireless.Wireless.get_organization_wireless_devices_wireless_controllers_by_device,
  - Paths used are
    get /organizations/{organizationId}/wireless/devices/wirelessControllers/byDevice,
"""

EXAMPLES = r"""
- name: Get all organizations _wireless _devices _wireless _controllers _by _device
  cisco.meraki.organizations_wireless_devices_wireless_controllers_by_device_info:
    meraki_api_key: "{{ meraki_api_key }}"
    meraki_base_url: "{{ meraki_base_url }}"
    meraki_single_request_timeout: "{{ meraki_single_request_timeout }}"
    meraki_certificate_path: "{{ meraki_certificate_path }}"
    meraki_requests_proxy: "{{ meraki_requests_proxy }}"
    meraki_wait_on_rate_limit: "{{ meraki_wait_on_rate_limit }}"
    meraki_nginx_429_retry_wait_time: "{{ meraki_nginx_429_retry_wait_time }}"
    meraki_action_batch_retry_wait_time: "{{ meraki_action_batch_retry_wait_time }}"
    meraki_retry_4xx_error: "{{ meraki_retry_4xx_error }}"
    meraki_retry_4xx_error_wait_time: "{{ meraki_retry_4xx_error_wait_time }}"
    meraki_maximum_retries: "{{ meraki_maximum_retries }}"
    meraki_output_log: "{{ meraki_output_log }}"
    meraki_log_file_prefix: "{{ meraki_log_file_prefix }}"
    meraki_log_path: "{{ meraki_log_path }}"
    meraki_print_console: "{{ meraki_print_console }}"
    meraki_suppress_logging: "{{ meraki_suppress_logging }}"
    meraki_simulate: "{{ meraki_simulate }}"
    meraki_be_geo_id: "{{ meraki_be_geo_id }}"
    meraki_caller: "{{ meraki_caller }}"
    meraki_use_iterator_for_get_pages: "{{ meraki_use_iterator_for_get_pages }}"
    meraki_inherit_logging_config: "{{ meraki_inherit_logging_config }}"
    networkIds: []
    serials: []
    controllerSerials: []
    perPage: 0
    startingAfter: string
    endingBefore: string
    organizationId: string
    total_pages: -1
    direction: next
  register: result
"""
RETURN = r"""
meraki_response:
  description: A dictionary or list with the response returned by the Cisco Meraki Python SDK
  returned: always
  type: dict
  sample: >
    {
      "items": [
        {
          "network": {
            "id": "string"
          },
          "serial": "string",
          "controller": {
            "serial": "string"
          },
          "joinedAt": "string",
          "model": "string",
          "tags": [
            {
              "policy": "string",
              "site": "string",
              "rf": "string"
            }
          ],
          "mode": "string",
          "countryCode": "string",
          "details": [
            {
              "name": "string",
              "value": "string"
            }
          ]
        }
      ],
      "meta": {
        "counts": {
          "items": {
            "total": 0,
            "remaining": 0
          }
        }
      }
    }
"""



#!/usr/bin/python
#
# Scaleway Load-balancer management module
#
# Copyright (C) 2018 Online SAS.
# https://www.scaleway.com
#
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: scaleway_lb
short_description: Scaleway load-balancer management module
author: Remy Leone (@remyleone)
description:
  - This module manages load-balancers on Scaleway.
extends_documentation_fragment:
  - community.general._scaleway
  - community.general._attributes
  - community.general._scaleway.actiongroup_scaleway

attributes:
  check_mode:
    support: full
  diff_mode:
    support: none
  action_group:
    version_added: 11.3.0

options:

  name:
    type: str
    description:
      - Name of the load-balancer.
    required: true

  description:
    type: str
    description:
      - Description of the load-balancer.
    required: true

  organization_id:
    type: str
    description:
      - Organization identifier.
    required: true

  state:
    type: str
    description:
      - Indicate desired state of the instance.
    default: present
    choices:
      - present
      - absent

  region:
    type: str
    description:
      - Scaleway zone.
    required: true
    choices:
      - nl-ams
      - fr-par
      - pl-waw

  tags:
    type: list
    elements: str
    default: []
    description:
      - List of tags to apply to the load-balancer.
  wait:
    description:
      - Wait for the load-balancer to reach its desired state before returning.
    type: bool
    default: false

  wait_timeout:
    type: int
    description:
      - Time to wait for the load-balancer to reach the expected state.
    default: 300

  wait_sleep_time:
    type: int
    description:
      - Time to wait before every attempt to check the state of the load-balancer.
    default: 3
"""

EXAMPLES = r"""
- name: Create a load-balancer
  community.general.scaleway_lb:
    name: foobar
    state: present
    organization_id: 951df375-e094-4d26-97c1-ba548eeb9c42
    region: fr-par
    tags:
      - hello

- name: Delete a load-balancer
  community.general.scaleway_lb:
    name: foobar
    state: absent
    organization_id: 951df375-e094-4d26-97c1-ba548eeb9c42
    region: fr-par
"""

RETURN = r"""
scaleway_lb:
  description: The load-balancer object.
  returned: success
  type: dict
  sample:
    {
      "backend_count": 0,
      "frontend_count": 0,
      "description": "Description of my load-balancer",
      "id": "00000000-0000-0000-0000-000000000000",
      "instances": [
        {
          "id": "00000000-0000-0000-0000-000000000000",
          "ip_address": "10.0.0.1",
          "region": "fr-par",
          "status": "ready"
        },
        {
          "id": "00000000-0000-0000-0000-000000000000",
          "ip_address": "10.0.0.2",
          "region": "fr-par",
          "status": "ready"
        }
      ],
      "ip": [
        {
          "id": "00000000-0000-0000-0000-000000000000",
          "ip_address": "192.168.0.1",
          "lb_id": "00000000-0000-0000-0000-000000000000",
          "region": "fr-par",
          "organization_id": "00000000-0000-0000-0000-000000000000",
          "reverse": ""
        }
      ],
      "name": "lb_ansible_test",
      "organization_id": "00000000-0000-0000-0000-000000000000",
      "region": "fr-par",
      "status": "ready",
      "tags": [
        "first_tag",
        "second_tag"
      ]
    }
"""

import datetime
import time

from ansible.module_utils.basic import AnsibleModule

from ansible_collections.community.general.plugins.module_utils._datetime import now
from ansible_collections.community.general.plugins.module_utils._scaleway import (
    SCALEWAY_ENDPOINT,
    SCALEWAY_REGIONS,
    Scaleway,
    scaleway_argument_spec,
)

STABLE_STATES = ("ready", "absent")

MUTABLE_ATTRIBUTES = ("name", "description")


def payload_from_wished_lb(wished_lb):
    return {
        "organization_id": wished_lb["organization_id"],
        "name": wished_lb["name"],
        "tags": wished_lb["tags"],
        "description": wished_lb["description"],
    }


def fetch_state(api, lb):
    api.module.debug(f"fetch_state of load-balancer: {lb['id']}")
    response = api.get(path=f"{api.api_path}/{lb['id']}")

    if response.status_code == 404:
        return "absent"

    if not response.ok:
        msg = f"Error during state fetching: ({response.status_code}) {response.json}"
        api.module.fail_json(msg=msg)

    try:
        api.module.debug(f"Load-balancer {lb['id']} in state: {response.json['status']}")
        return response.json["status"]
    except KeyError:
        api.module.fail_json(msg=f"Could not fetch state in {response.json}")


def wait_to_complete_state_transition(api, lb, force_wait=False):
    wait = api.module.params["wait"]
    if not (wait or force_wait):
        return
    wait_timeout = api.module.params["wait_timeout"]
    wait_sleep_time = api.module.params["wait_sleep_time"]

    start = now()
    end = start + datetime.timedelta(seconds=wait_timeout)
    while now() < end:
        api.module.debug("We are going to wait for the load-balancer to finish its transition")
        state = fetch_state(api, lb)
        if state in STABLE_STATES:
            api.module.debug("It seems that the load-balancer is not in transition anymore.")
            api.module.debug(f"load-balancer in state: {fetch_state(api, lb)}")
            break
        time.sleep(wait_sleep_time)
    else:
        api.module.fail_json(msg="Server takes too long to finish its transition")


def lb_attributes_should_be_changed(target_lb, wished_lb):
    diff = {attr: wished_lb[attr] for attr in MUTABLE_ATTRIBUTES if target_lb[attr] != wished_lb[attr]}

    if diff:
        return {attr: wished_lb[attr] for attr in MUTABLE_ATTRIBUTES}

    return {}


def present_strategy(api, wished_lb):
    changed = False

    response = api.get(path=api.api_path)
    if not response.ok:
        api.module.fail_json(msg=f"Error getting load-balancers [{response.status_code}: {response.json['message']}]")

    lbs_list = response.json["lbs"]
    lb_lookup = {lb["name"]: lb for lb in lbs_list}

    if wished_lb["name"] not in lb_lookup.keys():
        changed = True
        if api.module.check_mode:
            return changed, {"status": "A load-balancer would be created."}

        # Create Load-balancer
        api.warn(payload_from_wished_lb(wished_lb))
        creation_response = api.post(path=api.api_path, data=payload_from_wished_lb(wished_lb))

        if not creation_response.ok:
            msg = f"Error during lb creation: {creation_response.info['msg']}: '{creation_response.json['message']}' ({creation_response.json})"
            api.module.fail_json(msg=msg)

        wait_to_complete_state_transition(api=api, lb=creation_response.json)
        response = api.get(path=f"{api.api_path}/{creation_response.json['id']}")
        return changed, response.json

    target_lb = lb_lookup[wished_lb["name"]]
    patch_payload = lb_attributes_should_be_changed(target_lb=target_lb, wished_lb=wished_lb)

    if not patch_payload:
        return changed, target_lb

    changed = True
    if api.module.check_mode:
        return changed, {"status": "Load-balancer attributes would be changed."}

    lb_patch_response = api.put(path=f"{api.api_path}/{target_lb['id']}", data=patch_payload)

    if not lb_patch_response.ok:
        api.module.fail_json(
            msg=f"Error during load-balancer attributes update: [{lb_patch_response.status_code}: {lb_patch_response.json['message']}]"
        )

    wait_to_complete_state_transition(api=api, lb=target_lb)
    return changed, lb_patch_response.json


def absent_strategy(api, wished_lb):
    response = api.get(path=api.api_path)
    changed = False

    status_code = response.status_code
    lbs_json = response.json
    lbs_list = lbs_json["lbs"]

    if not response.ok:
        api.module.fail_json(msg=f"Error getting load-balancers [{status_code}: {response.json['message']}]")

    lb_lookup = {lb["name"]: lb for lb in lbs_list}
    if wished_lb["name"] not in lb_lookup.keys():
        return changed, {}

    target_lb = lb_lookup[wished_lb["name"]]
    changed = True
    if api.module.check_mode:
        return changed, {"status": "Load-balancer would be destroyed"}

    wait_to_complete_state_transition(api=api, lb=target_lb, force_wait=True)
    response = api.delete(path=f"{api.api_path}/{target_lb['id']}")
    if not response.ok:
        api.module.fail_json(msg=f"Error deleting load-balancer [{response.status_code}: {response.json}]")

    wait_to_complete_state_transition(api=api, lb=target_lb)
    return changed, response.json


state_strategy = {"present": present_strategy, "absent": absent_strategy}


def core(module):
    region = module.params["region"]
    wished_load_balancer = {
        "state": module.params["state"],
        "name": module.params["name"],
        "description": module.params["description"],
        "tags": module.params["tags"],
        "organization_id": module.params["organization_id"],
    }
    module.params["api_url"] = SCALEWAY_ENDPOINT
    api = Scaleway(module=module)
    api.api_path = f"lb/v1/regions/{region}/lbs"

    changed, summary = state_strategy[wished_load_balancer["state"]](api=api, wished_lb=wished_load_balancer)
    module.exit_json(changed=changed, scaleway_lb=summary)


def main():
    argument_spec = scaleway_argument_spec()
    argument_spec.update(
        dict(
            name=dict(required=True),
            description=dict(required=True),
            region=dict(required=True, choices=SCALEWAY_REGIONS),
            state=dict(choices=list(state_strategy.keys()), default="present"),
            tags=dict(type="list", elements="str", default=[]),
            organization_id=dict(required=True),
            wait=dict(type="bool", default=False),
            wait_timeout=dict(type="int", default=300),
            wait_sleep_time=dict(type="int", default=3),
        )
    )
    module = AnsibleModule(
        argument_spec=argument_spec,
        supports_check_mode=True,
    )

    core(module)


if __name__ == "__main__":
    main()



#!/usr/bin/python
#
# Copyright (c) 2021, Jyrki Gadinger <nilsding@nilsding.org>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

# Make coding more python3-ish
from __future__ import annotations

DOCUMENTATION = r"""
module: one_template

short_description: Manages OpenNebula templates

version_added: 2.4.0

requirements:
  - pyone

description:
  - Manages OpenNebula templates.
attributes:
  check_mode:
    support: partial
    details:
      - Note that check mode always returns RV(ignore:changed=true) for existing templates, even if the template would not actually
        change.
  diff_mode:
    support: none

options:
  id:
    description:
      - A O(id) of the template you would like to manage. If not set then a new template is created with the given O(name).
    type: int
  name:
    description:
      - A O(name) of the template you would like to manage. If a template with the given name does not exist it is created,
        otherwise it is managed by this module.
    type: str
  template:
    description:
      - A string containing the template contents.
    type: str
  state:
    description:
      - V(present) - state that is used to manage the template.
      - V(absent) - delete the template.
    choices: ["present", "absent"]
    default: present
    type: str
  filter:
    description:
      - V(user_primary_group) - Resources belonging to the user's primary group.
      - V(user) - Resources belonging to the user.
      - V(all) - All resources.
      - V(user_groups) - Resources belonging to the user and any of his groups.
    choices: [user_primary_group, user, all, user_groups]
    default: user
    type: str
    version_added: 10.3.0

extends_documentation_fragment:
  - community.general._opennebula
  - community.general._attributes

author:
  - "Jyrki Gadinger (@nilsding)"
"""

EXAMPLES = r"""
- name: Fetch the TEMPLATE by id
  community.general.one_template:
    id: 6459
  register: result

- name: Print the TEMPLATE properties
  ansible.builtin.debug:
    var: result

- name: Fetch the TEMPLATE by name
  community.general.one_template:
    name: tf-prd-users-workerredis-p6379a
  register: result

- name: Create a new or update an existing TEMPLATE
  community.general.one_template:
    name: generic-opensuse
    template: |
      CONTEXT = [
        HOSTNAME = "generic-opensuse"
      ]
      CPU = "1"
      CUSTOM_ATTRIBUTE = ""
      DISK = [
        CACHE = "writeback",
        DEV_PREFIX = "sd",
        DISCARD = "unmap",
        IMAGE = "opensuse-leap-15.2",
        IMAGE_UNAME = "oneadmin",
        IO = "threads",
        SIZE = "" ]
      MEMORY = "2048"
      NIC = [
        MODEL = "virtio",
        NETWORK = "testnet",
        NETWORK_UNAME = "oneadmin" ]
      OS = [
        ARCH = "x86_64",
        BOOT = "disk0" ]
      SCHED_REQUIREMENTS = "CLUSTER_ID=\\"100\\""
      VCPU = "2"

- name: Delete the TEMPLATE by id
  community.general.one_template:
    id: 6459
    state: absent
"""

RETURN = r"""
id:
  description: Template ID.
  type: int
  returned: when O(state=present)
  sample: 153
name:
  description: Template name.
  type: str
  returned: when O(state=present)
  sample: app1
template:
  description: The parsed template.
  type: dict
  returned: when O(state=present)
group_id:
  description: Template's group ID.
  type: int
  returned: when O(state=present)
  sample: 1
group_name:
  description: Template's group name.
  type: str
  returned: when O(state=present)
  sample: one-users
owner_id:
  description: Template's owner ID.
  type: int
  returned: when O(state=present)
  sample: 143
owner_name:
  description: Template's owner name.
  type: str
  returned: when O(state=present)
  sample: ansible-test
"""


from ansible_collections.community.general.plugins.module_utils._opennebula import OpenNebulaModule


class TemplateModule(OpenNebulaModule):
    def __init__(self):
        argument_spec = dict(
            id=dict(type="int"),
            name=dict(type="str"),
            state=dict(type="str", choices=["present", "absent"], default="present"),
            template=dict(type="str"),
            filter=dict(type="str", choices=["user_primary_group", "user", "all", "user_groups"], default="user"),
        )

        mutually_exclusive = [["id", "name"]]

        required_one_of = [("id", "name")]

        required_if = [["state", "present", ["template"]]]

        OpenNebulaModule.__init__(
            self,
            argument_spec,
            supports_check_mode=True,
            mutually_exclusive=mutually_exclusive,
            required_one_of=required_one_of,
            required_if=required_if,
        )

    def run(self, one, module, result):
        params = module.params
        id = params.get("id")
        name = params.get("name")
        desired_state = params.get("state")
        template_data = params.get("template")
        filter = params.get("filter")

        self.result = {}

        template = self.get_template_instance(id, name, filter)
        needs_creation = False
        if not template and desired_state != "absent":
            if id:
                module.fail_json(msg=f"There is no template with id={id}")
            else:
                needs_creation = True

        if desired_state == "absent":
            self.result = self.delete_template(template)
        else:
            if needs_creation:
                self.result = self.create_template(name, template_data, filter)
            else:
                self.result = self.update_template(template, template_data, filter)

        self.exit()

    def get_template(self, predicate, filter):
        # filter was included, for discussions see:
        # Issue: https://github.com/ansible-collections/community.general/issues/9278
        # PR: https://github.com/ansible-collections/community.general/pull/9547
        # the other two parameters are used for pagination, -1 for both essentially means "return all"
        filter_values = {"user_primary_group": -4, "user": -3, "all": -2, "user_groups": -1}
        pool = self.one.templatepool.info(filter_values[filter], -1, -1)

        for template in pool.VMTEMPLATE:
            if predicate(template):
                return template

        return None

    def get_template_by_id(self, template_id, filter):
        return self.get_template(lambda template: template_id == template.ID, filter)

    def get_template_by_name(self, name, filter):
        return self.get_template(lambda template: name == template.NAME, filter)

    def get_template_instance(self, requested_id, requested_name, filter):
        if requested_id:
            return self.get_template_by_id(requested_id, filter)
        else:
            return self.get_template_by_name(requested_name, filter)

    def get_template_info(self, template):
        info = {
            "id": template.ID,
            "name": template.NAME,
            "template": template.TEMPLATE,
            "user_name": template.UNAME,
            "user_id": template.UID,
            "group_name": template.GNAME,
            "group_id": template.GID,
        }

        return info

    def create_template(self, name, template_data, filter):
        if not self.module.check_mode:
            self.one.template.allocate(f'NAME = "{name}"\n{template_data}')

        result = self.get_template_info(self.get_template_by_name(name, filter))
        result["changed"] = True

        return result

    def update_template(self, template, template_data, filter):
        if not self.module.check_mode:
            # 0 = replace the whole template
            self.one.template.update(template.ID, template_data, 0)

        result = self.get_template_info(self.get_template_by_id(template.ID, filter))
        if self.module.check_mode:
            # Unfortunately it is not easy to detect if the template would have changed, therefore always report a change here.
            result["changed"] = True
        else:
            # if the previous parsed template data is not equal to the updated one, this has changed
            result["changed"] = result["template"] != template.TEMPLATE

        return result

    def delete_template(self, template):
        if not template:
            return {"changed": False}

        if not self.module.check_mode:
            self.one.template.delete(template.ID)

        return {"changed": True}


def main():
    TemplateModule().run_module()


if __name__ == "__main__":
    main()



#!/usr/bin/python
#
# Copyright (c) 2015, Brian Coca <bcoca@ansible.com>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: svc
author:
  - Brian Coca (@bcoca)
short_description: Manage C(daemontools) services
description:
  - Controls C(daemontools) services on remote hosts using the C(svc) utility.
extends_documentation_fragment:
  - community.general._attributes
attributes:
  check_mode:
    support: full
  diff_mode:
    support: none
options:
  name:
    description:
      - Name of the service to manage.
    type: str
    required: true
  state:
    description:
      - V(started)/V(stopped) are idempotent actions that do not run commands unless necessary.
      - V(restarted) always bounces the svc (svc -t) and V(killed) always bounces the svc (svc -k).
      - V(reloaded) sends a sigusr1 (svc -1).
      - V(once) runs a normally downed svc once (svc -o), not really an idempotent operation.
    type: str
    choices: [killed, once, reloaded, restarted, started, stopped]
  downed:
    description:
      - Should a C(down) file exist or not, if it exists it disables auto startup. Defaults to V(false). Downed does not imply
        stopped.
    type: bool
  enabled:
    description:
      - Whether the service is enabled or not, if disabled it also implies O(state=stopped). Take note that a service can
        be enabled and downed (no auto restart).
    type: bool
  service_dir:
    description:
      - Directory C(svscan) watches for services.
    type: str
    default: /service
  service_src:
    description:
      - Directory where services are defined, the source of symlinks to O(service_dir).
    type: str
    default: /etc/service
"""

EXAMPLES = r"""
- name: Start svc dnscache, if not running
  community.general.svc:
    name: dnscache
    state: started

- name: Stop svc dnscache, if running
  community.general.svc:
    name: dnscache
    state: stopped

- name: Kill svc dnscache, in all cases
  community.general.svc:
    name: dnscache
    state: killed

- name: Restart svc dnscache, in all cases
  community.general.svc:
    name: dnscache
    state: restarted

- name: Reload svc dnscache, in all cases
  community.general.svc:
    name: dnscache
    state: reloaded

- name: Using alternative svc directory location
  community.general.svc:
    name: dnscache
    state: reloaded
    service_dir: /var/service
"""

import os
import re
import traceback

from ansible.module_utils.basic import AnsibleModule


def _load_dist_subclass(cls, *args, **kwargs):
    """
    Used for derivative implementations
    """
    subclass = None

    distro = kwargs["module"].params["distro"]

    # get the most specific superclass for this platform
    if distro is not None:
        for sc in cls.__subclasses__():
            if sc.distro is not None and sc.distro == distro:
                subclass = sc
    if subclass is None:
        subclass = cls

    return super(cls, subclass).__new__(subclass)


class Svc:
    """
    Main class that handles daemontools, can be subclassed and overridden in case
    we want to use a 'derivative' like encore, s6, etc
    """

    # def __new__(cls, *args, **kwargs):
    #    return _load_dist_subclass(cls, args, kwargs)

    def __init__(self, module):
        self.extra_paths = ["/command", "/usr/local/bin"]
        self.report_vars = ["state", "enabled", "downed", "svc_full", "src_full", "pid", "duration", "full_state"]

        self.module = module

        self.name = module.params["name"]
        self.service_dir = module.params["service_dir"]
        self.service_src = module.params["service_src"]
        self.enabled = None
        self.downed = None
        self.full_state = None
        self.state = None
        self.pid = None
        self.duration = None

        self.svc_cmd = module.get_bin_path("svc", opt_dirs=self.extra_paths)
        self.svstat_cmd = module.get_bin_path("svstat", opt_dirs=self.extra_paths)
        self.svc_full = f"{self.service_dir}/{self.name}"
        self.src_full = f"{self.service_src}/{self.name}"

        self.enabled = os.path.lexists(self.svc_full)
        if self.enabled:
            self.downed = os.path.lexists(f"{self.svc_full}/down")
            self.get_status()
        else:
            self.downed = os.path.lexists(f"{self.src_full}/down")
            self.state = "stopped"

    def enable(self):
        if os.path.exists(self.src_full):
            try:
                os.symlink(self.src_full, self.svc_full)
            except OSError as e:
                self.module.fail_json(path=self.src_full, msg=f"Error while linking: {e}")
        else:
            self.module.fail_json(msg=f"Could not find source for service to enable ({self.src_full}).")

    def disable(self):
        try:
            os.unlink(self.svc_full)
        except OSError as e:
            self.module.fail_json(path=self.svc_full, msg=f"Error while unlinking: {e}")
        self.execute_command([self.svc_cmd, "-dx", self.src_full])

        src_log = f"{self.src_full}/log"
        if os.path.exists(src_log):
            self.execute_command([self.svc_cmd, "-dx", src_log])

    def get_status(self):
        rc, out, err = self.execute_command([self.svstat_cmd, self.svc_full])

        if err is not None and err:
            self.full_state = self.state = err
        else:
            self.full_state = out

            m = re.search(r"\(pid (\d+)\)", out)
            if m:
                self.pid = m.group(1)

            m = re.search(r"(\d+) seconds", out)
            if m:
                self.duration = m.group(1)

            if re.search(" up ", out):
                self.state = "start"
            elif re.search(" down ", out):
                self.state = "stopp"
            else:
                self.state = "unknown"
                return

            if re.search(" want ", out):
                self.state += "ing"
            else:
                self.state += "ed"

    def start(self):
        return self.execute_command([self.svc_cmd, "-u", self.svc_full])

    def stopp(self):
        return self.stop()

    def stop(self):
        return self.execute_command([self.svc_cmd, "-d", self.svc_full])

    def once(self):
        return self.execute_command([self.svc_cmd, "-o", self.svc_full])

    def reload(self):
        return self.execute_command([self.svc_cmd, "-1", self.svc_full])

    def restart(self):
        return self.execute_command([self.svc_cmd, "-t", self.svc_full])

    def kill(self):
        return self.execute_command([self.svc_cmd, "-k", self.svc_full])

    def execute_command(self, cmd):
        try:
            rc, out, err = self.module.run_command(cmd)
        except Exception as e:
            self.module.fail_json(msg=f"failed to execute: {e}", exception=traceback.format_exc())
        return (rc, out, err)

    def report(self):
        self.get_status()
        states = {}
        for k in self.report_vars:
            states[k] = self.__dict__[k]
        return states


# ===========================================
# Main control flow


def main():
    module = AnsibleModule(
        argument_spec=dict(
            name=dict(type="str", required=True),
            state=dict(type="str", choices=["killed", "once", "reloaded", "restarted", "started", "stopped"]),
            enabled=dict(type="bool"),
            downed=dict(type="bool"),
            service_dir=dict(type="str", default="/service"),
            service_src=dict(type="str", default="/etc/service"),
        ),
        supports_check_mode=True,
    )

    module.run_command_environ_update = dict(LANGUAGE="C", LC_ALL="C")

    state = module.params["state"]
    enabled = module.params["enabled"]
    downed = module.params["downed"]

    svc = Svc(module)
    changed = False
    dummy_orig_state = svc.report()  # TODO - is this not needed?

    if enabled is not None and enabled != svc.enabled:
        changed = True
        if not module.check_mode:
            try:
                if enabled:
                    svc.enable()
                else:
                    svc.disable()
            except OSError as e:
                module.fail_json(msg=f"Could not change service link: {e}")

    if state is not None and state != svc.state:
        changed = True
        if not module.check_mode:
            getattr(svc, state[:-2])()

    if downed is not None and downed != svc.downed:
        changed = True
        if not module.check_mode:
            d_file = f"{svc.svc_full}/down"
            try:
                if downed:
                    open(d_file, "a").close()
                else:
                    os.unlink(d_file)
            except OSError as e:
                module.fail_json(msg=f"Could not change downed file: {e} ")

    module.exit_json(changed=changed, svc=svc.report())


if __name__ == "__main__":
    main()



#!/usr/bin/python

# Copyright (c) 2013, Patrick Callahan <pmc@patrickcallahan.com>
# based on
#     openbsd_pkg
#         Copyright (c) 2013
#         Patrik Lundin <patrik.lundin.swe@gmail.com>
#
#     yum
#         Copyright (c) 2012, Red Hat, Inc
#         Written by Seth Vidal <skvidal at fedoraproject.org>
#
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: zypper
author:
  - "Patrick Callahan (@dirtyharrycallahan)"
  - "Alexander Gubin (@alxgu)"
  - "Thomas O'Donnell (@andytom)"
  - "Robin Roth (@robinro)"
  - "Andrii Radyk (@AnderEnder)"
short_description: Manage packages on SUSE and openSUSE
description:
  - Manage packages on SUSE and openSUSE using the zypper and rpm tools.
  - Also supports transactional updates, by running zypper inside C(/sbin/transactional-update --continue --drop-if-no-change
    --quiet run).
extends_documentation_fragment:
  - community.general._attributes
attributes:
  check_mode:
    support: full
  diff_mode:
    support: full
options:
  name:
    description:
      - Package name V(name) or package specifier or a list of either.
      - Can include a version like V(name=1.0), V(name>3.4) or V(name<=2.7). If a version is given, V(oldpackage) is implied
        and zypper is allowed to update the package within the version range given.
      - You can also pass a URL or a local path to a rpm file.
      - When using O(state=latest), this can be V(*), which updates all installed packages.
    required: true
    aliases: ['pkg']
    type: list
    elements: str
  state:
    description:
      - V(present) makes sure the package is installed.
      - V(latest) makes sure the latest version of the package is installed.
      - V(absent) makes sure the specified package is not installed.
      - V(dist-upgrade) makes sure the latest version of all installed packages from all enabled repositories is installed.
      - When using V(dist-upgrade), O(name) should be V(*).
    choices: [present, latest, absent, dist-upgrade, installed, removed]
    default: "present"
    type: str
  type:
    description:
      - The type of package to be operated on.
    choices: [package, patch, pattern, product, srcpackage, application]
    default: "package"
    type: str
  extra_args_precommand:
    description:
      - Add additional global target options to C(zypper).
      - Options should be supplied in a single line as if given in the command line.
    type: str
  disable_gpg_check:
    description:
      - Whether to disable to GPG signature checking of the package signature being installed. Has an effect only if O(state)
        is V(present) or V(latest).
    default: false
    type: bool
  auto_import_keys:
    description:
      - Whether to automatically import new repository signing keys. Adds C(--gpg-auto-import-keys) option to I(zypper).
      - Is only used when installing.
    default: false
    type: bool
    version_added: 11.3.0
  disable_recommends:
    description:
      - Corresponds to the C(--no-recommends) option for I(zypper). Default behavior (V(true)) modifies zypper's default behavior;
        V(false) does install recommended packages.
    default: true
    type: bool
  force:
    description:
      - Adds C(--force) option to I(zypper). Allows to downgrade packages and change vendor or architecture.
    default: false
    type: bool
  force_resolution:
    description:
      - Adds C(--force-resolution) option to I(zypper). Allows to (un)install packages with conflicting requirements (resolver
        chooses a solution).
    default: false
    type: bool
    version_added: '0.2.0'
  update_cache:
    description:
      - Run the equivalent of C(zypper refresh) before the operation. Disabled in check mode.
    default: false
    type: bool
    aliases: ["refresh"]
  oldpackage:
    description:
      - Adds C(--oldpackage) option to I(zypper). Allows to downgrade packages with less side-effects than force. This is
        implied as soon as a version is specified as part of the package name.
    default: false
    type: bool
  extra_args:
    description:
      - Add additional options to C(zypper) command.
      - Options should be supplied in a single line as if given in the command line.
    type: str
  allow_vendor_change:
    type: bool
    default: false
    description:
      - Adds C(--allow_vendor_change) option to I(zypper) dist-upgrade command.
    version_added: '0.2.0'
  replacefiles:
    type: bool
    default: false
    description:
      - Adds C(--replacefiles) option to I(zypper) install/update command.
    version_added: '0.2.0'
  clean_deps:
    type: bool
    default: false
    description:
      - Adds C(--clean-deps) option to I(zypper) remove command.
    version_added: '4.6.0'
  simple_errors:
    type: bool
    default: false
    description:
      - When set to V(true), provide a simplified error output (parses only the C(<message>) tag text in the XML output).
    version_added: '10.2.0'
  quiet:
    type: bool
    default: true
    description:
      - Adds C(--quiet) option to I(zypper) install/update command.
    version_added: '10.2.0'
  skip_post_errors:
    type: bool
    default: false
    description:
      - When set to V(true), ignore I(zypper) return code 107 (post install script errors).
    version_added: '10.6.0'
notes:
  - When used with a C(loop:) each package is processed individually, it is much more efficient to pass the list directly
    to the O(name) option.
requirements:
  - "zypper >= 1.0  # included in openSUSE >= 11.1 or SUSE Linux Enterprise Server/Desktop >= 11.0"
  - python-xml
  - rpm
"""

EXAMPLES = r"""
- name: Install nmap
  community.general.zypper:
    name: nmap
    state: present

- name: Install apache2 with recommended packages
  community.general.zypper:
    name: apache2
    state: present
    disable_recommends: false

- name: Apply a given patch
  community.general.zypper:
    name: openSUSE-2016-128
    state: present
    type: patch

- name: Remove the nmap package
  community.general.zypper:
    name: nmap
    state: absent

- name: Install the nginx rpm from a remote repo
  community.general.zypper:
    name: 'http://nginx.org/packages/sles/12/x86_64/RPMS/nginx-1.8.0-1.sles12.ngx.x86_64.rpm'
    state: present

- name: Install local rpm file
  community.general.zypper:
    name: /tmp/fancy-software.rpm
    state: present

- name: Update all packages
  community.general.zypper:
    name: '*'
    state: latest

- name: Install latest packages but dump error messages in a simplified format
  community.general.zypper:
    name: '*'
    state: latest
    simple_errors: true
    quiet: false

- name: Apply all available patches
  community.general.zypper:
    name: '*'
    state: latest
    type: patch

- name: Perform a dist-upgrade with additional arguments
  community.general.zypper:
    name: '*'
    state: dist-upgrade
    allow_vendor_change: true
    extra_args: '--allow-arch-change'

- name: Perform a installation of nmap with the install option replacefiles
  community.general.zypper:
    name: 'nmap'
    state: latest
    replacefiles: true

- name: Refresh repositories and update package openssl
  community.general.zypper:
    name: openssl
    state: present
    update_cache: true

- name: "Install specific version (possible comparisons: <, >, <=, >=, =)"
  community.general.zypper:
    name: 'docker>=1.10'
    state: present

- name: Wait 20 seconds to acquire the lock before failing
  community.general.zypper:
    name: mosh
    state: present
  environment:
    ZYPP_LOCK_TIMEOUT: 20

- name: Install the package with post-install error without failing
  community.general.zypper:
    name: <package_with_post_install_error>
    state: present
    skip_post_errors: true
"""

import os.path
import re
import xml
from xml.dom.minidom import parseString as parseXML

# import module snippets
from ansible.module_utils.basic import AnsibleModule


class Package:
    def __init__(self, name, prefix, version):
        self.name = name
        self.prefix = prefix
        self.version = version
        self.shouldinstall = prefix == "+"

    def __str__(self):
        return self.prefix + self.name + self.version


def split_name_version(name):
    """splits of the package name and desired version

    example formats:
        - docker>=1.10
        - apache=2.4

    Allowed version specifiers: <, >, <=, >=, =
    Allowed version format: [0-9.-]*

    Also allows a prefix indicating remove "-", "~" or install "+"
    """

    prefix = ""
    if name[0] in ["-", "~", "+"]:
        prefix = name[0]
        name = name[1:]
    if prefix == "~":
        prefix = "-"

    version_check = re.compile("^(.*?)((?:<|>|<=|>=|=)[0-9.-]*)?$")
    try:
        reres = version_check.match(name)
        name, version = reres.groups()
        if version is None:
            version = ""
        return prefix, name, version
    except Exception:
        return prefix, name, ""


def get_want_state(names, remove=False):
    packages = []
    urls = []
    for name in names:
        if "://" in name or name.endswith(".rpm"):
            urls.append(name)
        else:
            prefix, pname, version = split_name_version(name)
            if prefix not in ["-", "+"]:
                if remove:
                    prefix = "-"
                else:
                    prefix = "+"
            packages.append(Package(pname, prefix, version))
    return packages, urls


def get_installed_state(m, packages):
    "get installed state of packages"

    cmd = get_cmd(m, "search")
    cmd.extend(["--match-exact", "--details", "--installed-only"])
    cmd.extend([p.name for p in packages])
    return parse_zypper_xml(m, cmd, fail_not_found=False)[0]


def parse_zypper_xml(m, cmd, fail_not_found=True, packages=None):
    rc, stdout, stderr = m.run_command(cmd, check_rc=False)

    try:
        dom = parseXML(stdout)
    except xml.parsers.expat.ExpatError as exc:
        m.fail_json(msg=f"Failed to parse zypper xml output: {exc}", rc=rc, stdout=stdout, stderr=stderr, cmd=cmd)

    if rc == 104:
        # exit code 104 is ZYPPER_EXIT_INF_CAP_NOT_FOUND (no packages found)
        if fail_not_found:
            errmsg = dom.getElementsByTagName("message")[-1].childNodes[0].data
            m.fail_json(msg=errmsg, rc=rc, stdout=stdout, stderr=stderr, cmd=cmd)
        else:
            return {}, rc, stdout, stderr
    elif rc in [0, 102, 103, 106, 107]:
        # zypper exit codes
        # 0: success
        # 106: signature verification failed
        # 102: ZYPPER_EXIT_INF_REBOOT_NEEDED - Returned after a successful installation of a patch which requires reboot of computer.
        # 103: zypper was upgraded, run same command again
        # 107: ZYPPER_EXIT_INF_RPM_SCRIPT_FAILED -  Some of the packages %post install scripts returned an error, but package is installed.
        if packages is None:
            firstrun = True
            packages = {}
        else:
            firstrun = False
        solvable_list = dom.getElementsByTagName("solvable")
        for solvable in solvable_list:
            name = solvable.getAttribute("name")
            packages[name] = {}
            packages[name]["version"] = solvable.getAttribute("edition")
            packages[name]["oldversion"] = solvable.getAttribute("edition-old")
            status = solvable.getAttribute("status")
            packages[name]["installed"] = status == "installed"
            packages[name]["group"] = solvable.parentNode.nodeName
        if rc == 103 and firstrun:
            # if this was the first run and it failed with 103
            # run zypper again with the same command to complete update
            return parse_zypper_xml(m, cmd, fail_not_found=fail_not_found, packages=packages)
        if rc == 107 and m.params["skip_post_errors"] and firstrun:
            # if this was the first run and it failed with 107 with skip_post_errors flag
            # run zypper again with the same command to complete update
            return parse_zypper_xml(m, cmd, fail_not_found=fail_not_found, packages=packages)

        # apply simple_errors logic to rc 0,102,103,106,107
        if m.params["simple_errors"]:
            stdout = get_simple_errors(dom) or stdout

        return packages, rc, stdout, stderr

    # apply simple_errors logic to rc other than 0,102,103,106,107
    if m.params["simple_errors"]:
        stdout = get_simple_errors(dom) or stdout

    m.fail_json(msg=f"Zypper run command failed with return code {rc}.", rc=rc, stdout=stdout, stderr=stderr, cmd=cmd)


def get_simple_errors(dom):
    simple_errors = []
    message_xml_tags = dom.getElementsByTagName("message")

    if message_xml_tags is None:
        return None

    for x in message_xml_tags:
        simple_errors.append(x.firstChild.data)
    return " \n".join(simple_errors)


def get_cmd(m, subcommand):
    "puts together the basic zypper command arguments with those passed to the module"
    is_install = subcommand in ["install", "update", "patch", "dist-upgrade"]
    is_refresh = subcommand == "refresh"
    cmd = [m.get_bin_path("zypper", required=True), "--non-interactive", "--xmlout"]
    if m.params["quiet"]:
        cmd.append("--quiet")
    if transactional_updates():
        cmd = [
            m.get_bin_path("transactional-update", required=True),
            "--continue",
            "--drop-if-no-change",
            "--quiet",
            "run",
        ] + cmd
    if m.params["extra_args_precommand"]:
        args_list = m.params["extra_args_precommand"].split()
        cmd.extend(args_list)
    # add global options before zypper command
    if (is_install or is_refresh) and m.params["disable_gpg_check"]:
        cmd.append("--no-gpg-checks")
    if is_install and m.params["auto_import_keys"]:
        cmd.append("--gpg-auto-import-keys")

    if subcommand == "search":
        cmd.append("--disable-repositories")

    cmd.append(subcommand)
    if subcommand not in ["patch", "dist-upgrade"] and not is_refresh:
        cmd.extend(["--type", m.params["type"]])
    if m.check_mode and subcommand != "search":
        cmd.append("--dry-run")
    if is_install:
        cmd.append("--auto-agree-with-licenses")
        if m.params["disable_recommends"]:
            cmd.append("--no-recommends")
        if m.params["force"]:
            cmd.append("--force")
        if m.params["force_resolution"]:
            cmd.append("--force-resolution")
        if m.params["oldpackage"]:
            cmd.append("--oldpackage")
        if m.params["replacefiles"]:
            cmd.append("--replacefiles")
    if subcommand == "remove":
        if m.params["clean_deps"]:
            cmd.append("--clean-deps")
    if subcommand == "dist-upgrade" and m.params["allow_vendor_change"]:
        cmd.append("--allow-vendor-change")
    if m.params["extra_args"]:
        args_list = m.params["extra_args"].split(" ")
        cmd.extend(args_list)

    return cmd


def set_diff(m, retvals, result):
    # TODO: if there is only one package, set before/after to version numbers
    packages = {"installed": [], "removed": [], "upgraded": []}
    if result:
        for p in result:
            group = result[p]["group"]
            if group == "to-upgrade":
                versions = f" ({result[p]['oldversion']} => {result[p]['version']})"
                packages["upgraded"].append(p + versions)
            elif group == "to-install":
                packages["installed"].append(p)
            elif group == "to-remove":
                packages["removed"].append(p)

    output = ""
    for state in packages:
        if packages[state]:
            output += f"{state}: {', '.join(packages[state])}\n"
    if "diff" not in retvals:
        retvals["diff"] = {}
    if "prepared" not in retvals["diff"]:
        retvals["diff"]["prepared"] = output
    else:
        retvals["diff"]["prepared"] += f"\n{output}"


def package_present(m, name, want_latest):
    "install and update (if want_latest) the packages in name_install, while removing the packages in name_remove"
    retvals = {"rc": 0, "stdout": "", "stderr": ""}
    packages, urls = get_want_state(name)

    # add oldpackage flag when a version is given to allow downgrades
    if any(p.version for p in packages):
        m.params["oldpackage"] = True

    if not want_latest:
        # for state=present: filter out already installed packages
        # if a version is given leave the package in to let zypper handle the version
        # resolution
        packageswithoutversion = [p for p in packages if not p.version]
        prerun_state = {}
        if packageswithoutversion:
            prerun_state = get_installed_state(m, packageswithoutversion)
        # generate lists of packages to install or remove
        packages = [p for p in packages if p.shouldinstall != (p.name in prerun_state)]

    if not packages and not urls:
        # nothing to install/remove and nothing to update
        return None, retvals

    # zypper install also updates packages
    cmd = get_cmd(m, "install")
    cmd.append("--")
    cmd.extend(urls)
    # pass packages to zypper
    # allow for + or - prefixes in install/remove lists
    # also add version specifier if given
    # do this in one zypper run to allow for dependency-resolution
    # for example "-exim postfix" runs without removing packages depending on mailserver
    cmd.extend([str(p) for p in packages])

    retvals["cmd"] = cmd
    result, retvals["rc"], retvals["stdout"], retvals["stderr"] = parse_zypper_xml(m, cmd)

    return result, retvals


def package_update_all(m):
    "run update or patch on all available packages"

    retvals = {"rc": 0, "stdout": "", "stderr": ""}
    if m.params["type"] == "patch":
        cmdname = "patch"
    elif m.params["state"] == "dist-upgrade":
        cmdname = "dist-upgrade"
    else:
        cmdname = "update"

    cmd = get_cmd(m, cmdname)
    retvals["cmd"] = cmd
    result, retvals["rc"], retvals["stdout"], retvals["stderr"] = parse_zypper_xml(m, cmd)
    return result, retvals


def package_absent(m, name):
    "remove the packages in name"
    retvals = {"rc": 0, "stdout": "", "stderr": ""}
    # Get package state
    packages, urls = get_want_state(name, remove=True)
    if any(p.prefix == "+" for p in packages):
        m.fail_json(msg="Can not combine '+' prefix with state=remove/absent.")
    if urls:
        m.fail_json(msg="Can not remove via URL.")
    if m.params["type"] == "patch":
        m.fail_json(msg="Can not remove patches.")
    prerun_state = get_installed_state(m, packages)
    packages = [p for p in packages if p.name in prerun_state]

    if not packages:
        return None, retvals

    cmd = get_cmd(m, "remove")
    cmd.extend([p.name + p.version for p in packages])

    retvals["cmd"] = cmd
    result, retvals["rc"], retvals["stdout"], retvals["stderr"] = parse_zypper_xml(m, cmd)
    return result, retvals


def repo_refresh(m):
    "update the repositories"
    retvals = {"rc": 0, "stdout": "", "stderr": ""}

    cmd = get_cmd(m, "refresh")

    retvals["cmd"] = cmd
    result, retvals["rc"], retvals["stdout"], retvals["stderr"] = parse_zypper_xml(m, cmd)

    return retvals


def get_fs_type_and_readonly_state(mount_point):
    with open("/proc/mounts") as file:
        for line in file.readlines():
            fields = line.split()
            path = fields[1]
            if path == mount_point:
                fs = fields[2]
                opts = fields[3]
                return fs, "ro" in opts.split(",")
    return None


def transactional_updates():
    return os.path.exists("/usr/sbin/transactional-update") and get_fs_type_and_readonly_state("/") == ("btrfs", True)


# ===========================================
# Main control flow


def main():
    module = AnsibleModule(
        argument_spec=dict(
            name=dict(required=True, aliases=["pkg"], type="list", elements="str"),
            state=dict(
                default="present", choices=["absent", "installed", "latest", "present", "removed", "dist-upgrade"]
            ),
            type=dict(
                default="package", choices=["package", "patch", "pattern", "product", "srcpackage", "application"]
            ),
            extra_args_precommand=dict(),
            disable_gpg_check=dict(default=False, type="bool"),
            auto_import_keys=dict(default=False, type="bool"),
            disable_recommends=dict(default=True, type="bool"),
            force=dict(default=False, type="bool"),
            force_resolution=dict(default=False, type="bool"),
            update_cache=dict(aliases=["refresh"], default=False, type="bool"),
            oldpackage=dict(default=False, type="bool"),
            extra_args=dict(),
            allow_vendor_change=dict(default=False, type="bool"),
            replacefiles=dict(default=False, type="bool"),
            clean_deps=dict(default=False, type="bool"),
            simple_errors=dict(default=False, type="bool"),
            quiet=dict(default=True, type="bool"),
            skip_post_errors=dict(default=False, type="bool"),
        ),
        supports_check_mode=True,
    )

    module.run_command_environ_update = dict(LANGUAGE="C", LC_ALL="C")

    name = module.params["name"]
    state = module.params["state"]
    update_cache = module.params["update_cache"]

    # remove empty strings from package list
    name = [_f for _f in name if _f]

    # Refresh repositories
    if update_cache and not module.check_mode:
        retvals = repo_refresh(module)

        if retvals["rc"] != 0:
            module.fail_json(msg="Zypper refresh run failed.", **retvals)

    # Perform requested action
    if name == ["*"] and state in ["latest", "dist-upgrade"]:
        packages_changed, retvals = package_update_all(module)
    elif name != ["*"] and state == "dist-upgrade":
        module.fail_json(msg="Can not dist-upgrade specific packages.")
    else:
        if state in ["absent", "removed"]:
            packages_changed, retvals = package_absent(module, name)
        elif state in ["installed", "present", "latest"]:
            packages_changed, retvals = package_present(module, name, state == "latest")

    retvals["changed"] = retvals["rc"] in [0, 102] and bool(packages_changed)

    if module._diff:
        set_diff(module, retvals, packages_changed)

    if retvals["rc"] not in [0, 102]:
        module.fail_json(msg="Zypper run failed.", **retvals)

    if not retvals["changed"]:
        del retvals["stdout"]
        del retvals["stderr"]

    module.exit_json(name=name, state=state, update_cache=update_cache, **retvals)


if __name__ == "__main__":
    main()
